import pytest

from hugr import ops, tys, val
from hugr.build.cond_loop import Conditional, ConditionalError, TailLoop
from hugr.build.dfg import Dfg
from hugr.package import Package
from hugr.std.int import INT_T, IntVal

from .conftest import QUANTUM_EXT, H, Measure, validate

EITHER_T = tys.Either([tys.Qubit], [tys.Qubit, INT_T])


def build_cond(h: Conditional) -> None:
    with pytest.raises(ConditionalError, match="Case 2 out of possible range."):
        h.add_case(2)

    with h.add_case(0) as case0:
        q, b = case0.inputs()
        case0.set_outputs(q, b)

    with h.add_case(1) as case1:
        q, _i, b = case1.inputs()
        case1.set_outputs(q, b)


def test_cond() -> None:
    h = Conditional(EITHER_T, [tys.Bool])
    build_cond(h)
    validate(h.hugr)


def test_nested_cond() -> None:
    h = Dfg(tys.Qubit)
    (q,) = h.inputs()
    tagged_q = h.add(ops.Left(EITHER_T)(q))

    with h.add_conditional(tagged_q, h.load(val.TRUE)) as cond:
        build_cond(cond)

    h.set_outputs(*cond[:2])
    validate(h.hugr)

    # build then insert
    con = Conditional(EITHER_T, [tys.Bool])
    build_cond(con)

    h = Dfg(tys.Qubit)
    (q,) = h.inputs()
    tagged_q = h.add(ops.Left(EITHER_T)(q))
    cond_n = h.insert_conditional(con, tagged_q, h.load(val.TRUE))
    h.set_outputs(*cond_n[:2])
    validate(h.hugr)


def test_if_else() -> None:
    # apply an H if a bool is true.
    h = Dfg(tys.Qubit)
    (q,) = h.inputs()
    with h.add_if(h.load(val.TRUE), q) as if_:
        if_.set_outputs(if_.add(H(if_.input_node[0])))

    with if_.add_else() as else_:
        else_.set_outputs(else_.input_node[0])

    h.set_outputs(else_.conditional_node)

    validate(Package([h.hugr], [QUANTUM_EXT]))


def test_incomplete() -> None:
    def _build_incomplete():
        with Conditional(EITHER_T, [tys.Bool]) as c, c.add_case(0) as case0:
            q, b = case0.inputs()
            case0.set_outputs(q, b)

    with pytest.raises(
        ConditionalError, match="All cases must be added before exiting context."
    ):
        _build_incomplete()


def test_tail_loop() -> None:
    # apply H while measure is true
    def build_tl(tl: TailLoop) -> None:
        q, b = tl.add(Measure(tl.add(H(tl.input_node[0]))))[:]

        tl.set_loop_outputs(b, q)

    h = Dfg(tys.Qubit)
    (q,) = h.inputs()

    with h.add_tail_loop([], [q]) as tl:
        build_tl(tl)
    h.set_outputs(tl)

    validate(Package([h.hugr], [QUANTUM_EXT]))

    # build then insert
    tl = TailLoop([], [tys.Qubit])
    build_tl(tl)

    h = Dfg(tys.Qubit)
    (q,) = h.inputs()
    tl_n = h.insert_tail_loop(tl, [q], [])
    h.set_outputs(tl_n)
    validate(Package([h.hugr], [QUANTUM_EXT]))


def test_complex_tail_loop() -> None:
    h = Dfg(tys.Qubit)
    (q,) = h.inputs()

    # loop passes qubit to itself, and a bool as in-out
    with h.add_tail_loop([q], [h.load(val.TRUE)]) as tl:
        q, b = tl.inputs()

        # if b is true, return first variant (just qubit)
        with tl.add_if(b, q) as if_:
            (q,) = if_.inputs()
            tagged_q = if_.add(ops.Continue(EITHER_T)(q))
            if_.set_outputs(tagged_q)

        # else return second variant (qubit, int)
        with if_.add_else() as else_:
            (q,) = else_.inputs()
            tagged_q_i = else_.add(ops.Break(EITHER_T)(q, else_.load(IntVal(1))))
            else_.set_outputs(tagged_q_i)

        # finish with Sum output from if-else, and bool from inputs
        tl.set_loop_outputs(else_.conditional_node, b)

    # loop returns [qubit, int, bool]
    h.set_outputs(*tl[:3])

    validate(h.hugr)


def test_conditional_bug() -> None:
    # bug with case ordering https://github.com/CQCL/hugr/issues/1596
    cond = Conditional(tys.Either([tys.USize()], [tys.Unit]), [])
    with cond.add_case(1) as case:
        case.set_outputs()
    with cond.add_case(0) as case:
        case.set_outputs()
    validate(cond.hugr)



"""Core data structures for HUGR."""

from __future__ import annotations

import json
from collections.abc import Iterable, Iterator, Mapping
from dataclasses import dataclass, field
from queue import Queue
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import deprecated

import hugr.model as model
import hugr.ops as ops
from hugr._serialization.ops import OpType as SerialOp
from hugr._serialization.serial_hugr import SerialHugr
from hugr.envelope import (
    EnvelopeConfig,
    make_envelope,
    make_envelope_str,
    read_envelope_hugr,
    read_envelope_hugr_str,
)
from hugr.exceptions import ParentBeforeChild
from hugr.ops import (
    CFG,
    Call,
    Conditional,
    Const,
    Custom,
    DataflowOp,
    FuncDefn,
    IncompleteOp,
    Module,
    Op,
)
from hugr.tys import Kind, Type, ValueKind
from hugr.utils import BiMap
from hugr.val import Value

from .node_port import (
    Direction,
    InPort,
    Node,
    NodeIdx,
    OutPort,
    PortOffset,
    ToNode,
    _SubPort,
)

if TYPE_CHECKING:
    import graphviz as gv  # type: ignore[import-untyped]

    from hugr import ext
    from hugr.val import Value

    from .render import RenderConfig


@dataclass()
class NodeData:
    """Node weights in HUGR graph. Defined by an operation and parent node."""

    #: The operation of the node.
    op: Op
    #: The parent node, or None for the root node.
    parent: Node | None
    _num_inps: int = field(default=0, repr=False)
    _num_outs: int = field(default=0, repr=False)
    children: list[Node] = field(default_factory=list, repr=False)
    metadata: dict[str, Any] = field(default_factory=dict)

    def _to_serial(self, node: Node) -> SerialOp:
        o = self.op._to_serial(self.parent if self.parent else node)

        return SerialOp(root=o)  # type: ignore[arg-type]


_SO = _SubPort[OutPort]
_SI = _SubPort[InPort]

P = TypeVar("P", InPort, OutPort)
K = TypeVar("K", InPort, OutPort)
OpVar = TypeVar("OpVar", bound=Op)
OpVarCov = TypeVar("OpVarCov", bound=Op, covariant=True)


@dataclass()
class Hugr(Mapping[Node, NodeData], Generic[OpVarCov]):
    """The core HUGR datastructure.

    Args:
        root_op: The operation for the root node. Defaults to a Module.

    Examples:
        >>> h = Hugr()
        >>> h.entrypoint_op()
        Module()
        >>> h[h.entrypoint].op
        Module()
        >>> dfg_h = Hugr(ops.DFG([tys.Bool]))
        >>> dfg_h[dfg_h.entrypoint].op
        DFG(inputs=[Bool])
    """

    # The module at the root of the HUGR.
    module_root: Node
    # Entrypoint node for the HUGR.
    #
    # Most traversals and rewrite operations start from this node.
    #
    # This node may be of any optype that's the parent to a region, and is a
    # descendant of the module definition at the HUGR root (or the root itself).
    entrypoint: Node
    # List of nodes, with None for deleted nodes.
    _nodes: list[NodeData | None]
    # Bidirectional map of links between ports.
    _links: BiMap[_SO, _SI]
    # List of free node indices, populated when nodes are deleted.
    _free_nodes: list[Node]

    def __init__(self, entrypoint_op: OpVarCov | None = None) -> None:
        self._free_nodes = []
        self._links = BiMap()
        self._nodes = []
        self.entrypoint = Node(0)
        self.module_root = Node(0)

        # The root of a HUGR is always a module.
        self.module_root = self._add_node(Module(), None, 0)
        self.entrypoint = self.module_root

        unsupported_op_msg = (
            f"Creating new HUGRs with entrypoint {entrypoint_op} is not supported"
        )

        # Depending on the entrypoint op, we may need to
        # wrap nest it inside the root module.
        match entrypoint_op:
            case None | Module():
                pass
            case ops.FuncDefn():
                self.entrypoint = self.add_node(entrypoint_op, self.module_root)
            case _:
                from hugr.build import Function

                # Some operations are unsupported, as they require additional context to
                # be valid (e.g. cfg blocks, case statements, etc.).
                if not ops.is_dataflow_op(entrypoint_op):
                    raise ValueError(unsupported_op_msg)
                # Explicit type required to keep mypy happy
                df_op: ops.DataflowOp = entrypoint_op

                # Only region containers are allowed to be entrypoints
                match df_op:
                    case CFG() | Conditional():
                        pass
                    case _ if ops.is_df_parent_op(df_op):
                        pass
                    case _:
                        raise ValueError(unsupported_op_msg)

                inputs, outputs = None, None
                try:
                    sig = df_op.outer_signature()
                    inputs = sig.input
                    outputs = sig.output
                except IncompleteOp:
                    match df_op:
                        case CFG():
                            inputs = df_op.inputs
                        case _:
                            inputs = df_op._inputs()

                parent_op = FuncDefn("main", inputs, [])
                func = Function.new_nested(parent_op, self, self.module_root)

                if outputs is not None:
                    self.entrypoint = func.add_op(df_op, *func.inputs())
                    func.set_outputs(*self.entrypoint.outputs())
                else:
                    # Connecting the entrypoint to the function's output is delayed
                    # until `set_outputs` is called.
                    # See `hugr._connect_df_entrypoint_outputs`.
                    self.entrypoint = self.add_node(df_op, func)
                    func._wire_up(self.entrypoint, func.inputs())
                    df_op._entrypoint_requires_wiring = True

    def __getitem__(self, key: ToNode) -> NodeData:
        key = key.to_node()
        try:
            n = self._nodes[key.idx]
        except IndexError:
            n = None
        if n is None:
            raise KeyError(key)
        return n

    def __iter__(self) -> Iterator[Node]:
        return (
            Node(idx, data.metadata)
            for idx, data in enumerate(self._nodes)
            if data is not None
        )

    def __len__(self) -> int:
        return self.num_nodes()

    def _get_typed_op(self, node: ToNode, cl: type[OpVar]) -> OpVar:
        op = self[node].op
        assert isinstance(op, cl)
        return op

    def nodes(self) -> Iterable[tuple[Node, NodeData]]:
        """Iterator over all the nodes of the hugr and their data.

        To get the descendants of the entrypoint, use `descendants()`.
        """
        return self.items()

    def sorted_region_nodes(self, parent: Node) -> Iterator[Node]:
        """Iterator over a topological ordering of all the hugr nodes.

        Note that the sort is performed within a hugr region and non-local
        edges are ignored.

        Args:
            parent: The parent node of the region to sort.

        Raises:
            ValueError: If the region contains a cycle.

        Examples:
            >>> from hugr.build.tracked_dfg import TrackedDfg
            >>> from hugr.std.logic import Not
            >>> dfg = TrackedDfg(tys.Bool)
            >>> [b] = dfg.track_inputs()
            >>> for _ in range(6):
            ...     _= dfg.add(Not(b));
            >>> dfg.set_tracked_outputs()
            >>> nodes = list(dfg.hugr)
            >>> list(dfg.hugr.sorted_region_nodes(nodes[4]))
            [Node(5), Node(7), Node(8), Node(9), Node(10), Node(11), Node(12), Node(6)]
        """
        # A dict to keep track of how many times we see a node.
        # Store the Nodes with the input degrees as values.
        # Implementation uses Kahn's algorithm
        # https://en.wikipedia.org/wiki/Topological_sorting#Kahn's_algorithm
        visit_dict: dict[Node, int] = {}
        queue: Queue[Node] = Queue()
        for node in self.children(parent):
            incoming = 0
            for n in self.input_neighbours(node):
                same_region = self[n].parent == parent
                # Only update the degree of the node if edge is within the same region.
                # We do not count non-local edges.
                if same_region:
                    incoming += 1
            if incoming:
                visit_dict[node] = incoming
            # If a Node has no dependencies, add it to the queue.
            else:
                queue.put(node)

        while not queue.empty():
            new_node = queue.get()
            yield new_node

            for neigh in self.output_neighbours(new_node):
                visit_dict[neigh] -= 1
                if visit_dict[neigh] == 0:
                    del visit_dict[neigh]
                    queue.put(neigh)

        # If our dict is non-empty here then our graph contains a cycle
        if visit_dict:
            err = "Graph contains a cycle. No topological ordering exists."
            raise ValueError(err)

    def links(self) -> Iterator[tuple[OutPort, InPort]]:
        """Iterator over all the links in the HUGR.

        Returns:
            Iterator of pairs of outgoing port and the incoming ports.
        """
        return ((src.port, tgt.port) for src, tgt in self._links.items())

    def children(self, node: ToNode | None = None) -> list[Node]:
        """The child nodes of a given `node`.

        Args:
            node: Parent node. Defaults to the HUGR entrypoint.

        Returns:
            List of child nodes.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_node(ops.Const(val.TRUE))
            >>> h.children(h.entrypoint)
            [Node(1)]
        """
        node = node or self.entrypoint
        return self[node].children

    def descendants(self, node: ToNode | None = None) -> Iterable[Node]:
        """Iterator over all the descendants of the hugr entrypoint.

        Traverses the HUGR graph in a breadth-first manner, starting from
        the entrypoint.

        To get all the nodes in the HUGR, use `nodes()`.

        Args:
            node: Parent node. Defaults to the HUGR entrypoint.

        Returns:
            List of child nodes.
        """
        queue: Queue[Node] = Queue()
        queue.put(node.to_node() if node is not None else self.entrypoint)
        while not queue.empty():
            node = queue.get()
            yield node
            for child in self.children(node):
                queue.put(child)

    def _add_node(
        self,
        op: Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Node:
        parent = parent.to_node() if parent else None
        node_data = NodeData(op, parent, metadata=metadata or {})

        if self._free_nodes:
            node = self._free_nodes.pop()
            self._nodes[node.idx] = node_data
        else:
            node = Node(len(self._nodes), {})
            self._nodes.append(node_data)
        node._num_out_ports = num_outs
        node._metadata = node_data.metadata
        if parent:
            self[parent].children.append(node)

        self._update_node_outs(node, num_outs)
        return node

    def _update_node_outs(self, node: Node, num_outs: int | None) -> Node:
        """Update the number of outgoing ports for a node.

        Returns:
            The updated node.
        """
        return self._update_port_count(node, num_outs=num_outs)

    def _update_port_count(
        self, node: Node, *, num_inps: int | None = None, num_outs: int | None
    ) -> Node:
        """Update the number of incoming and outgoing ports for a node.

        If `num_inps` or `num_outs` is None, the corresponding count is not updated.

        Returns:
            The updated node.
        """
        if num_inps is None and num_outs is None:
            return node

        if num_inps is not None:
            self[node]._num_inps = num_inps
        if num_outs is not None:
            self[node]._num_outs = num_outs
            node._num_out_ports = num_outs

        if node.idx == self.entrypoint.idx:
            self.entrypoint = node
        if node.idx == self.module_root.idx:
            self.module_root = node

        return node

    def add_node(
        self,
        op: Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Node:
        """Add a node to the HUGR.

        Args:
            op: Operation of the node.
            parent: Parent node of added node. Defaults to HUGR entrypoint if None.
            num_outs: Number of output ports expected for this node. Defaults to None.
            metadata: A dictionary of metadata to associate with the node.
                Defaults to None.

        Returns:
            Handle to the added node.
        """
        parent = parent or self.entrypoint
        return self._add_node(op, parent, num_outs, metadata)

    def add_const(
        self,
        value: Value,
        parent: ToNode | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Node:
        """Add a constant node to the HUGR.

        Args:
            value: Value of the constant.
            parent: Parent node of added node. Defaults to HUGR entrypoint if None.
            metadata: A dictionary of metadata to associate with the node.
                Defaults to None.

        Returns:
            Handle to the added node.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> h[n].op
            Const(TRUE)
        """
        return self.add_node(Const(value), parent, metadata=metadata)

    def delete_node(self, node: ToNode) -> NodeData | None:
        """Delete a node from the HUGR.

        Args:
            node: Node to delete.

        Returns:
            The deleted node data, or None if the node was not found.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> deleted = h.delete_node(n)
            >>> deleted.op
            Const(TRUE)
            >>> len(h)
            1
        """
        node = node.to_node()
        parent = self[node].parent
        if parent:
            self[parent].children.remove(node)
        for inp, _ in self.incoming_links(node):
            self._links.delete_right(_SubPort(inp))
        for out, _ in self.outgoing_links(node):
            self._links.delete_left(_SubPort(out))

        weight, self._nodes[node.idx] = self._nodes[node.idx], None

        # Free up the metadata dictionary
        node._metadata = {}

        self._free_nodes.append(node)
        return weight

    def _unused_sub_offset(self, port: P) -> _SubPort[P]:
        d: dict[_SO, _SI] | dict[_SI, _SO]
        match port:
            case OutPort(_):
                d = self._links.fwd
            case InPort(_):
                d = self._links.bck
        sub_port = _SubPort(port)
        while sub_port in d:
            sub_port = sub_port.next_sub_offset()
        return sub_port

    def has_link(self, src: OutPort, dst: InPort) -> bool:
        """Check if there is a link between two ports.

        Args:
            src: Source port.
            dst: Destination port.

        Returns:
            True if there is a link, False otherwise.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.has_link(df.input_node.out(0), df.output_node.inp(0))
            True
        """
        return dst in self.linked_ports(src)

    def add_link(self, src: OutPort, dst: InPort) -> None:
        """Add a link (edge) between two nodes to the HUGR,
          from an outgoing port to an incoming port.

        Args:
            src: Source port.
            dst: Destination port.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> list(df.hugr.linked_ports(df.input_node[0]))
            [InPort(Node(6), 0)]
        """
        src_sub = self._unused_sub_offset(src)
        dst_sub = self._unused_sub_offset(dst)
        # if self._links.get_left(dst_sub) is not None:
        #     dst = replace(dst, _sub_offset=dst._sub_offset + 1)
        self._links.insert_left(src_sub, dst_sub)

        self[src.node]._num_outs = max(self[src.node]._num_outs, src.offset + 1)
        self[dst.node]._num_inps = max(self[dst.node]._num_inps, dst.offset + 1)

    def add_order_link(self, src: ToNode, dst: ToNode) -> None:
        """Add a state order link between two nodes.

        Args:
            src: Source node.
            dst: Destination node.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_order_link(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_order_links(df.input_node))
            [Node(6)]
        """
        source = src.out(-1)
        target = dst.inp(-1)
        if not self.has_link(source, target):
            self.add_link(source, target)

    def delete_link(self, src: OutPort, dst: InPort) -> None:
        """Delete a link (edge) between two nodes from the HUGR.

        Args:
            src: Source port.
            dst: Destination port.
        """
        try:
            sub_offset = next(
                i for i, inp in enumerate(self.linked_ports(src)) if inp == dst
            )
            self._links.delete_left(_SubPort(src, sub_offset))
        except StopIteration:
            return
        # TODO make sure sub-offset is handled correctly

    def entrypoint_op(self) -> OpVarCov:
        """The operation of the root node.

        Examples:
            >>> h = Hugr()
            >>> h.entrypoint_op()
            Module()
        """
        return cast("OpVarCov", self[self.entrypoint].op)

    def num_nodes(self) -> int:
        """The number of nodes in the HUGR.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> h.num_nodes()
            2
        """
        return len(self._nodes) - len(self._free_nodes)

    def num_ports(self, node: ToNode, direction: Direction) -> int:
        """The number of ports of a node in a given direction.
        Not necessarily the number of connected ports - if port `i` is
        connected, then all ports `0..i` are assumed to exist.

        Args:
            node: Node to query.
            direction: Direction of ports to count.

        Examples:
            >>> h = Hugr()
            >>> n1 = h.add_const(val.TRUE)
            >>> n2 = h.add_const(val.FALSE)
            >>> h.add_link(n1.out(0), n2.inp(2)) # not a valid link!
            >>> h.num_ports(n1, Direction.OUTGOING)
            1
            >>> h.num_ports(n2, Direction.INCOMING)
            3
        """
        return (
            self.num_in_ports(node)
            if direction == Direction.INCOMING
            else self.num_out_ports(node)
        )

    def num_in_ports(self, node: ToNode) -> int:
        """The number of incoming ports of a node. See :meth:`num_ports`."""
        return self[node]._num_inps

    def num_out_ports(self, node: ToNode) -> int:
        """The number of outgoing ports of a node. See :meth:`num_ports`."""
        return self[node]._num_outs

    def _linked_ports(
        self, port: P, links: dict[_SubPort[P], _SubPort[K]]
    ) -> Iterable[K]:
        sub_port = _SubPort(port)
        while sub_port in links:
            # sub offset not used in API
            yield links[sub_port].port
            sub_port = sub_port.next_sub_offset()

    @overload
    def linked_ports(self, port: OutPort) -> Iterable[InPort]: ...
    @overload
    def linked_ports(self, port: InPort) -> Iterable[OutPort]: ...
    def linked_ports(self, port: OutPort | InPort):
        """Return an iterable of In(Out)Ports linked to given Out(In)Port.

        Args:
            port: Given port.

        Returns:
            Iterator over linked ports.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.set_outputs(df.input_node[0])
            >>> list(df.hugr.linked_ports(df.input_node[0]))
            [InPort(Node(6), 0)]

        """
        match port:
            case OutPort(_):
                return self._linked_ports(port, self._links.fwd)
            case InPort(_):
                return self._linked_ports(port, self._links.bck)

    # TODO: single linked port

    def outgoing_order_links(self, node: ToNode) -> Iterable[Node]:
        """Iterator over nodes connected by an outgoing state order link from a
        given node.

        Args:
            node: Source node of state order link.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_order_links(df.input_node))
            [Node(6)]
        """
        return (p.node for p in self.linked_ports(node.out(-1)))

    def incoming_order_links(self, node: ToNode) -> Iterable[Node]:
        """Iterator over nodes connected by an incoming state order link to a
        given node.

        Args:
            node: Destination node of state order link.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.incoming_order_links(df.output_node))
            [Node(5)]
        """
        return (p.node for p in self.linked_ports(node.inp(-1)))

    def _node_links(
        self, node: ToNode, links: dict[_SubPort[P], _SubPort[K]]
    ) -> Iterable[tuple[P, list[K]]]:
        try:
            direction = next(iter(links.keys())).port.direction
        except StopIteration:
            return
        # iterate over known offsets
        for offset in range(self.num_ports(node, direction)):
            port = cast("P", node.port(offset, direction))
            yield port, list(self._linked_ports(port, links))

    def outgoing_links(self, node: ToNode) -> Iterable[tuple[OutPort, list[InPort]]]:
        """Iterator over outgoing links from a given node.

        Args:
            node: Node to query.

        Returns:
            Iterator of pairs of outgoing port and the incoming ports connected
            to that port.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.outgoing_links(df.input_node))
            [(OutPort(Node(5), 0), [InPort(Node(6), 0), InPort(Node(6), 1)])]
        """
        return self._node_links(node, self._links.fwd)

    def incoming_links(self, node: ToNode) -> Iterable[tuple[InPort, list[OutPort]]]:
        """Iterator over incoming links to a given node.

        Args:
            node: Node to query.

        Returns:
            Iterator of pairs of incoming port and the outgoing ports connected
            to that port.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.incoming_links(df.output_node))
            [(InPort(Node(6), 0), [OutPort(Node(5), 0)]), (InPort(Node(6), 1), [OutPort(Node(5), 0)])]
        """  # noqa: E501
        return self._node_links(node, self._links.bck)

    def neighbours(
        self, node: ToNode, direction: Direction | None = None
    ) -> Iterable[Node]:
        """Iterator over the neighbours of a node.

        Args:
            node: Node to query.
            direction: If given, only return neighbours in that direction.

        Returns:
            Iterator of nodes connected to `node`, ordered by direction and port
            offset. Nodes connected via multiple links will be returned multiple times.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.neighbours(df.input_node))
            [Node(6), Node(6)]
            >>> list(df.hugr.neighbours(df.output_node, Direction.OUTGOING))
            []
        """
        if direction is None or direction == Direction.INCOMING:
            for _, linked_outputs in self.incoming_links(node):
                for out_port in linked_outputs:
                    yield out_port.node
        if direction is None or direction == Direction.OUTGOING:
            for _, linked_inputs in self.outgoing_links(node):
                for in_port in linked_inputs:
                    yield in_port.node

    def input_neighbours(self, node: ToNode) -> Iterable[Node]:
        """Iterator over the input neighbours of a node.

        Args:
            node: Node to query.

        Returns:
            Iterator of nodes connected to `node` via incoming links.
            Nodes connected via multiple links will be returned multiple times.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.input_neighbours(df.output_node))
            [Node(5), Node(5)]
        """
        return self.neighbours(node, Direction.INCOMING)

    def output_neighbours(self, node: ToNode) -> Iterable[Node]:
        """Iterator over the output neighbours of a node.

        Args:
            node: Node to query.

        Returns:
            Iterator of nodes connected to `node` via outgoing links.
            Nodes connected via multiple links will be returned multiple times.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.output_neighbours(df.input_node))
            [Node(6), Node(6)]
        """
        return self.neighbours(node, Direction.OUTGOING)

    def num_incoming(self, node: Node) -> int:
        """The number of incoming links to a `node`.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.num_incoming(df.output_node)
            1
        """
        return sum(1 for _ in self.incoming_links(node))

    def num_outgoing(self, node: ToNode) -> int:
        """The number of outgoing links from a `node`.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.num_outgoing(df.input_node)
            1
        """
        return sum(1 for _ in self.outgoing_links(node))

    # TODO: num_links and _linked_ports

    def port_kind(self, port: InPort | OutPort) -> Kind:
        """The kind of a `port`.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.port_kind(df.input_node.out(0))
            ValueKind(Bool)
        """
        return self[port.node].op.port_kind(port)

    def port_type(self, port: InPort | OutPort) -> Type | None:
        """The type of a `port`, if the kind is
        :class:`ValueKind <hugr.tys.ValueKind>`, else None.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.port_type(df.input_node.out(0))
            Bool
        """
        op = self[port.node].op
        if isinstance(op, DataflowOp):
            return op.port_type(port)
        if isinstance(op, Call) and isinstance(port, OutPort):
            kind = self.port_kind(port)
            if isinstance(kind, ValueKind):
                return kind.ty
        return None

    def insert_hugr(self, hugr: Hugr, parent: ToNode | None = None) -> dict[Node, Node]:
        """Insert a HUGR entrypoint and all its descendants into this HUGR.

        If the inserted HUGR entrypoint was not its module root, some nodes will
        be ignored.

        Args:
            hugr: HUGR to insert.
            parent: Parent for root of inserted HUGR. Defaults to None.

        Returns:
            Mapping from node indices in inserted HUGR to their new indices
            in this HUGR.

        Examples:
            >>> d = dfg.Dfg()
            >>> h = Hugr()
            >>> h.insert_hugr(d.hugr)
            {Node(4): Node(1), Node(5): Node(2), Node(6): Node(3)}
        """
        mapping: dict[Node, Node] = {}

        for node in hugr.descendants():
            node_data = hugr[node]
            # relies on parents being inserted before any children
            if node == hugr.entrypoint:
                node_parent = parent
            else:
                if node_data.parent not in mapping:
                    raise ParentBeforeChild
                node_parent = mapping[node_data.parent]
            mapping[node] = self.add_node(
                node_data.op,
                node_parent,
                num_outs=node_data._num_outs,
                metadata=node_data.metadata,
            )

        for src, dst in hugr._links.items():
            # Neighbours that are not descendants of the entrypoint are ignored.
            if src.port.node not in mapping or dst.port.node not in mapping:
                continue
            self.add_link(
                mapping[src.port.node].out(src.port.offset),
                mapping[dst.port.node].inp(dst.port.offset),
            )
        return mapping

    def _to_serial(self) -> SerialHugr:
        """Serialize the HUGR."""

        def _serialize_link(
            link: tuple[_SO, _SI],
        ) -> tuple[tuple[NodeIdx, PortOffset], tuple[NodeIdx, PortOffset]]:
            src, dst = link
            s, d = self._constrain_offset(src.port), self._constrain_offset(dst.port)
            return (src.port.node.idx, s), (dst.port.node.idx, d)

        nodes: list[SerialOp] = []
        metadata = []
        entrypoint = 0
        for node_idx, data in enumerate(self._nodes):
            if data is None:
                continue
            node = Node(node_idx)
            serial_idx = len(nodes)

            # non contiguous indices will be erased
            nodes.append(data._to_serial(Node(serial_idx, {})))
            metadata.append(data.metadata if data.metadata else None)
            if self.entrypoint == node:
                entrypoint = serial_idx

        return SerialHugr(
            nodes=nodes,
            edges=[_serialize_link(link) for link in self._links.items()],
            metadata=metadata,
            entrypoint=entrypoint,
        )

    def _constrain_offset(self, p: P) -> PortOffset:
        # An offset of -1 is a special case, indicating an order edge,
        # not counted in the number of ports.
        if p.offset < 0:
            assert p.offset == -1, "Only order edges are allowed with offset < 0"
            offset = self.num_ports(p.node, p.direction)
        else:
            offset = p.offset

        return offset

    def resolve_extensions(self, registry: ext.ExtensionRegistry) -> Hugr:
        """Resolve extension types and operations in the HUGR by matching them to
        extensions in the registry.
        """
        for node in self:
            op = self[node].op
            if isinstance(op, Custom):
                self[node].op = op.resolve(registry)
        return self

    def _connect_df_entrypoint_outputs(self) -> None:
        """If this Hugr was created by wrapping a dataflow operation entrypoint in a
        function, connect the entrypoint outputs to the function outputs.

        See `hugr.__init__` for more details.
        """
        from hugr.build import Function

        if not isinstance(self.entrypoint_op(), DataflowOp):
            return

        func_node = self[self.entrypoint].parent
        assert func_node is not None, "Only module entrypoints may be HUGR roots"
        func_op = self[func_node].op
        if not isinstance(func_op, FuncDefn) or func_op._outputs is not None:
            return

        func = Function._new_existing(self, func_node)
        func.set_outputs(*self.entrypoint.outputs())

    @classmethod
    def _from_serial(cls, serial: SerialHugr) -> Hugr:
        """Load a HUGR from a serialized form."""
        assert serial.nodes, "The encoded Hugr is empty"

        def get_meta(idx: int) -> dict[str, Any]:
            if not serial.metadata:
                return {}
            if idx < len(serial.metadata):
                return serial.metadata[idx] or {}
            return {}

        # The first node is always the HUGR root.
        root_node = serial.nodes[0]
        assert (
            root_node.root.parent == 0
        ), "The encoded Hugr root must be the first node"
        root_op = root_node.root.deserialize()
        hugr = Hugr(root_op)
        hugr[hugr.entrypoint].metadata = get_meta(0)
        # In older versions, this may not be a module operation. If that's the
        # case, we rely on hugr's initializer to wrap the root in a module
        # and ignore the boilerplate nodes.
        boilerplate_nodes = hugr.num_nodes() - 1

        for idx, serial_node in enumerate(serial.nodes):
            if idx == 0:
                continue

            node_meta = get_meta(idx)
            parent: Node | None = Node(serial_node.root.parent)

            serial_node.root.parent = -1
            n = hugr._add_node(
                serial_node.root.deserialize(), parent, metadata=node_meta
            )
            assert (
                n.idx == idx + boilerplate_nodes
            ), "Nodes should be added contiguously"

            if idx == serial.entrypoint:
                hugr.entrypoint = n

        for (src_node, src_offset), (dst_node, dst_offset) in serial.edges:
            if src_offset is None or dst_offset is None:
                continue
            hugr.add_link(
                Node(src_node, _metadata=get_meta(src_node)).out(src_offset),
                Node(dst_node, _metadata=get_meta(dst_node)).inp(dst_offset),
            )

        return hugr

    @staticmethod
    def from_bytes(envelope: bytes) -> Hugr:
        """Deserialize a byte string to a Hugr object.

        Some envelope formats can be read from a string. See :meth:`from_str`.

        Args:
            envelope: The byte string representing a Hugr envelope.

        Returns:
            The deserialized Hugr object.

        Raises:
            ValueError: If the envelope does not contain exactly one module.
        """
        return read_envelope_hugr(envelope)

    @staticmethod
    def from_str(envelope: str) -> Hugr:
        """Deserialize a string to a Hugr object.

        Not all envelope formats can be read from a string.
        See :meth:`from_bytes` for a more general method.

        Args:
            envelope: The string representing a Hugr envelope.

        Returns:
            The deserialized Hugr object.

        Raises:
            ValueError: If the envelope does not contain exactly one module.
        """
        return read_envelope_hugr_str(envelope)

    def to_bytes(self, config: EnvelopeConfig | None = None) -> bytes:
        """Serialize the HUGR into an envelope byte string.

        Some envelope formats can be encoded into a string. See :meth:`to_str`.
        """
        config = config or EnvelopeConfig.BINARY
        return make_envelope(self, config)

    def to_str(self, config: EnvelopeConfig | None = None) -> str:
        """Serialize the package to a HUGR envelope string.

        Not all envelope formats can be encoded into a string.
        See :meth:`to_bytes` for a more general method.
        """
        config = config or EnvelopeConfig.TEXT
        return make_envelope_str(self, config)

    @deprecated("Use HUGR envelopes instead. See the `to_bytes` and `to_str` methods.")
    def to_json(self) -> str:
        """Serialize the HUGR to a JSON string.

        For most use cases, it is recommended to store a HUGR package instead.
        See :meth:`hugr.package.Package.to_bytes`.
        """
        return self._to_serial().to_json()

    def to_model(self) -> model.Module:
        """Export this module into the hugr model format."""
        from hugr.model.export import ModelExport

        export = ModelExport(self)
        region = export.export_region_module(self.module_root)
        return model.Module(region)

    @classmethod
    @deprecated("Use HUGR envelopes instead. See the `to_bytes` and `to_str` methods.")
    def load_json(cls, json_str: str) -> Hugr:
        """Deserialize a JSON string into a HUGR.

        For most use cases, it is recommended to use package serialization instead.
        See :meth:`hugr.package.Package.from_bytes`.
        """
        json_dict = json.loads(json_str)
        serial = SerialHugr.load_json(json_dict)
        return cls._from_serial(serial)

    def render_dot(self, config: RenderConfig | None = None) -> gv.Digraph:
        """Render the HUGR to a graphviz Digraph.

        Args:
            config: Render configuration.

        Returns:
            The graphviz Digraph.
        """
        from .render import DotRenderer

        return DotRenderer(config).render(self)

    def store_dot(
        self, filename: str, format: str = "svg", config: RenderConfig | None = None
    ) -> None:
        """Render the HUGR to a graphviz dot file.

        Args:
            filename: The file to render to.
            format: The format used for rendering ('pdf', 'png', etc.).
                Defaults to SVG.
            config: Render configuration.
        """
        from .render import DotRenderer

        DotRenderer(config).store(self, filename=filename, format=format)



from hugr import tys
from hugr.build.function import Module
from hugr.envelope import EnvelopeConfig, EnvelopeFormat
from hugr.package import Package


def test_envelope():
    mod = Module()
    f_id = mod.define_function("id", [tys.Qubit])
    f_id.set_outputs(f_id.input_node[0])

    mod2 = Module()
    f_id_decl = mod2.declare_function(
        "id", tys.PolyFuncType([], tys.FunctionType([tys.Qubit], [tys.Qubit]))
    )
    f_main = mod2.define_main([tys.Qubit])
    q = f_main.input_node[0]
    call = f_main.call(f_id_decl, q)
    f_main.set_outputs(call)
    package = Package([mod.hugr, mod2.hugr])

    # Binary compression roundtrip
    for format in [EnvelopeFormat.JSON]:
        for compression in [None, 0]:
            encoded = package.to_bytes(EnvelopeConfig(format=format, zstd=compression))
            decoded = Package.from_bytes(encoded)
            assert decoded == package

    # String roundtrip
    encoded = package.to_str(EnvelopeConfig.TEXT)
    decoded = Package.from_str(encoded)
    assert decoded == package



"""Builder classes for defining functions and modules in HUGR."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hugr import ops
from hugr.build.dfg import DefinitionBuilder, Function
from hugr.hugr import Hugr

if TYPE_CHECKING:
    from hugr.hugr.node_port import Node
    from hugr.tys import PolyFuncType, TypeBound, TypeRow

__all__ = ["Function", "Module"]


@dataclass
class Module(DefinitionBuilder[ops.Module]):
    """Build a top-level HUGR module.

    Examples:
        >>> m = Module()
        >>> m.hugr.entrypoint_op()
        Module()
    """

    hugr: Hugr[ops.Module]

    def __init__(self) -> None:
        self.hugr = Hugr(ops.Module())

    def define_main(self, input_types: TypeRow) -> Function:
        """Define the 'main' function in the module. See :meth:`define_function`."""
        return self.define_function("main", input_types)

    def declare_function(self, name: str, signature: PolyFuncType) -> Node:
        """Add a function declaration to the module.

        Args:
            name: The name of the function.
            signature: The (polymorphic) signature of the function.

        Returns:
            The node representing the function declaration.

        Examples:
            >>> from hugr.function import Module
            >>> m = Module()
            >>> sig = tys.PolyFuncType([], tys.FunctionType.empty())
            >>> m.declare_function("f", sig)
            Node(1)
        """
        return self.hugr.add_node(ops.FuncDecl(name, signature), self.hugr.entrypoint)

    def add_alias_decl(self, name: str, bound: TypeBound) -> Node:
        """Add a type alias declaration."""
        return self.hugr.add_node(ops.AliasDecl(name, bound), self.hugr.entrypoint)

    @property
    def metadata(self) -> dict[str, object]:
        """Metadata associated with this module."""
        return self.hugr.entrypoint.metadata



"""Shared utility classes and functions."""

from collections.abc import Hashable, ItemsView, Iterable, Mapping, MutableMapping
from dataclasses import dataclass, field
from typing import Generic, Protocol, TypeVar

L = TypeVar("L", bound=Hashable)
R = TypeVar("R", bound=Hashable)


class NotBijection(Exception):
    """Initial map is not a bijection."""


@dataclass()
class BiMap(MutableMapping, Generic[L, R]):
    """Bidirectional map backed by two dictionaries, between left types `L` and
    right types `R`.
    """

    fwd: dict[L, R] = field(default_factory=dict)
    bck: dict[R, L] = field(default_factory=dict)

    def __init__(self, fwd: Mapping[L, R] | None = None) -> None:
        """Initialize a bidirectional map.

        Args:
            fwd: Left to right mapping. Defaults to empty.

        Raises:
            NotBijection: If the initial map is not a bijection.
        """
        fwd = fwd or {}
        if len(fwd) != len(set(fwd.values())):
            raise NotBijection
        self.fwd = dict(fwd)
        self.bck = {v: k for k, v in fwd.items()}

    def __getitem__(self, key: L) -> R:
        """Get the right value for a left key.

        Args:
            key: Left key.

        Raises:
            KeyError: If the key is not found.

        Example:
            >>> bm = BiMap({"a": 1})
            >>> bm["a"]
            1
        """
        return self.fwd[key]

    def __setitem__(self, key: L, value: R) -> None:
        """See :meth:`insert_left`."""
        self.insert_left(key, value)

    def __delitem__(self, key: L) -> None:
        """See :meth:`delete_left`."""
        self.delete_left(key)

    def __iter__(self):
        return iter(self.fwd)

    def __len__(self) -> int:
        return len(self.fwd)

    def items(self) -> ItemsView[L, R]:
        """Iterator over left, right pairs.

        Example:
            >>> bm = BiMap({"a": 1, "b": 2})
            >>> list(bm.items())
            [('a', 1), ('b', 2)]
        """
        return self.fwd.items()

    def get_left(self, key: R) -> L | None:
        """Get a left value using a right key.

        Example:
            >>> bm = BiMap({"a": 1})
            >>> bm.get_left(1)
            'a'
            >>> bm.get_left(2)
        """
        return self.bck.get(key)

    def get_right(self, key: L) -> R | None:
        """Get a right value using a left key.

        Example:
            >>> bm = BiMap({"a": 1})
            >>> bm.get_right("a")
            1
            >>> bm.get_right("b")
        """
        return self.fwd.get(key)

    def insert_left(self, key: L, value: R) -> None:
        """Insert a left key and right value.
        If the key or value already exist, the existing key-value pair is replaced.

        Args:
            key: Left key.
            value: Right value.

        Example:
            >>> bm = BiMap()
            >>> bm.insert_left("a", 1)
            >>> bm["a"]
            1
        """
        if (existing_key := self.bck.get(value)) is not None:
            del self.fwd[existing_key]
        if (existing_value := self.fwd.get(key)) is not None:
            del self.bck[existing_value]
        self.fwd[key] = value
        self.bck[value] = key

    def insert_right(self, key: R, value: L) -> None:
        """Insert a right key and left value.
        If the key or value already exist, the existing key-value pair is replaced.

        Args:
            key: Right key.
            value: Left value.

        Example:
            >>> bm = BiMap()
            >>> bm.insert_right(1, "a")
            >>> bm["a"]
            1
        """
        self.insert_left(value, key)

    def delete_left(self, key: L) -> None:
        """Delete a left key and its right value.

        Args:
            key: Left key.

        Raises:
            KeyError: If the key is not found.

        Example:
            >>> bm = BiMap({"a": 1})
            >>> bm.delete_left("a")
            >>> bm
            BiMap({})
        """
        del self.bck[self.fwd[key]]
        del self.fwd[key]

    def delete_right(self, key: R) -> None:
        """Delete a right key and its left value.

        Args:
            key: Right key.

        Raises:
            KeyError: If the key is not found.

        Example:
            >>> bm = BiMap({"a": 1})
            >>> bm.delete_right(1)
            >>> bm
            BiMap({})
        """
        del self.fwd[self.bck[key]]
        del self.bck[key]

    def __repr__(self) -> str:
        return f"BiMap({self.fwd})"


S = TypeVar("S", covariant=True)


class SerCollection(Protocol[S]):
    """Protocol for serializable objects."""

    def _to_serial_root(self) -> S:
        """Convert to serializable root model."""
        ...  # pragma: no cover


class DeserCollection(Protocol[S]):
    """Protocol for deserializable objects."""

    def deserialize(self) -> S:
        """Deserialize from model."""
        ...  # pragma: no cover


def ser_it(it: Iterable[SerCollection[S]]) -> list[S]:
    """Serialize an iterable of serializable objects."""
    return [v._to_serial_root() for v in it]


def deser_it(it: Iterable[DeserCollection[S]]) -> list[S]:
    """Deserialize an iterable of deserializable objects."""
    return [v.deserialize() for v in it]


T = TypeVar("T")


def comma_sep_str(items: Iterable[T]) -> str:
    """Join items with commas and str."""
    return ", ".join(map(str, items))


def comma_sep_repr(items: Iterable[T]) -> str:
    """Join items with commas and repr."""
    return ", ".join(map(repr, items))



"""HUGR package and pointed package interfaces."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generic, TypeVar, cast

from typing_extensions import deprecated

import hugr._serialization.extension as ext_s
import hugr.model as model
from hugr.envelope import (
    EnvelopeConfig,
    make_envelope,
    make_envelope_str,
    read_envelope,
    read_envelope_str,
)
from hugr.ops import FuncDecl, FuncDefn, Op

if TYPE_CHECKING:
    from hugr.ext import Extension
    from hugr.hugr.base import Hugr
    from hugr.hugr.node_port import Node

__all__ = [
    "Package",
    "PackagePointer",
    "ModulePointer",
    "ExtensionPointer",
    "NodePointer",
    "FuncDeclPointer",
    "FuncDefnPointer",
]


@dataclass(frozen=True)
class Package:
    """A package of HUGR modules and extensions.


    The HUGRs may refer to the included extensions or those not included.
    """

    #: HUGR modules in the package.
    modules: list[Hugr]
    #: Extensions included in the package.
    extensions: list[Extension] = field(default_factory=list)

    def _to_serial(self) -> ext_s.Package:
        return ext_s.Package(
            modules=[m._to_serial() for m in self.modules],
            extensions=[e._to_serial() for e in self.extensions],
        )

    @staticmethod
    def from_bytes(envelope: bytes) -> Package:
        """Deserialize a byte string to a Package object.

        Some envelope formats can be read from a string. See :meth:`from_str`.

        Args:
            envelope: The byte string representing a Package.

        Returns:
            The deserialized Package object.
        """
        return read_envelope(envelope)

    @staticmethod
    def from_str(envelope: str) -> Package:
        """Deserialize a string to a Package object.

        Not all envelope formats can be read from a string.
        See :meth:`from_bytes` for a more general method.

        Args:
            envelope: The string representing a Package.

        Returns:
            The deserialized Package object.
        """
        return read_envelope_str(envelope)

    def to_bytes(self, config: EnvelopeConfig | None = None) -> bytes:
        """Serialize the package to a HUGR envelope byte string.

        Some envelope formats can be encoded into a string. See :meth:`to_str`.
        """
        config = config or EnvelopeConfig.BINARY
        return make_envelope(self, config)

    def to_str(self, config: EnvelopeConfig | None = None) -> str:
        """Serialize the package to a HUGR envelope string.

        Not all envelope formats can be encoded into a string.
        See :meth:`to_bytes` for a more general method.
        """
        config = config or EnvelopeConfig.TEXT
        return make_envelope_str(self, config)

    @deprecated("Use HUGR envelopes instead. See the `to_bytes` and `to_str` methods.")
    def to_json(self) -> str:
        """Serialize the package to a printable HUGR envelope string."""
        return self._to_serial().model_dump_json()

    @classmethod
    @deprecated(
        "Use HUGR envelopes instead. See the `from_bytes` and `from_str` methods."
    )
    def from_json(cls, json_str: str) -> Package:
        """Deserialize a JSON string to a Package object.

        Args:
            json_str: The JSON string representing a Package.

        Returns:
            The deserialized Package object.
        """
        return ext_s.Package.model_validate_json(json_str).deserialize()

    def to_model(self) -> model.Package:
        """Export the package as its hugr model representation.

        At the moment this does not yet contain the extensions.
        """
        return model.Package([module.to_model() for module in self.modules])


@dataclass(frozen=True)
class PackagePointer:
    """Classes that point to packages and their inner contents."""

    #: Package pointed to.
    package: Package


@dataclass(frozen=True)
class ModulePointer(PackagePointer):
    """Pointer to a module in a package.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
    """

    #: Index of the module in the package.
    module_index: int

    @property
    def module(self) -> Hugr:
        """Hugr definition of the module."""
        return self.package.modules[self.module_index]

    def to_executable_package(self) -> ExecutablePackage:
        """Create an executable package from a module containing a main function.

        Raises:
            ValueError: If the module does not contain a main function.
        """
        module = self.module
        try:
            main_node = next(
                n
                for n in module.children()
                if isinstance((f_def := module[n].op), FuncDefn)
                and f_def.f_name == "main"
            )
        except StopIteration as e:
            msg = "Module does not contain a main function"
            raise ValueError(msg) from e
        return ExecutablePackage(self.package, self.module_index, main_node)


@dataclass(frozen=True)
class ExtensionPointer(PackagePointer):
    """Pointer to an extension in a package.

    Args:
        package: Package pointed to.
        extension_index: Index of the extension in the package.
    """

    #: Index of the extension in the package.
    extension_index: int

    @property
    def extension(self) -> Extension:
        """Extension definition."""
        return self.package.extensions[self.extension_index]


OpType = TypeVar("OpType", bound=Op)


@dataclass(frozen=True)
class NodePointer(Generic[OpType], ModulePointer):
    """Pointer to a node in a module.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
        node: Node pointed to
    """

    #: Node pointed to.
    node: Node

    @property
    def node_op(self) -> OpType:
        """Get the operation of the node."""
        return cast(OpType, self.module[self.node].op)


@dataclass(frozen=True)
class FuncDeclPointer(NodePointer[FuncDecl]):
    """Pointer to a function declaration in a module.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
        node: Node containing the function declaration.
    """

    @property
    def func_decl(self) -> FuncDecl:
        """Function declaration."""
        return self.node_op


@dataclass(frozen=True)
class FuncDefnPointer(NodePointer[FuncDefn]):
    """Pointer to a function definition in a module.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
        node: Node containing the function definition
    """

    @property
    def func_defn(self) -> FuncDefn:
        """Function definition."""
        return self.node_op


@dataclass(frozen=True)
class ExecutablePackage(FuncDefnPointer):
    """PackagePointer with a defined entrypoint node.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
        node: Node containing the entry point function definition.
    """

    @property
    def entry_point_node(self) -> Node:
        """Get the entry point node of the package."""
        return self.node



"""The main HUGR structure."""

from .base import Hugr, NodeData
from .node_port import (
    Direction,
    InPort,
    Node,
    OutPort,
    Wire,
)

__all__ = [
    "Hugr",
    "NodeData",
    "Direction",
    "InPort",
    "Wire",
    "OutPort",
    "Node",
]



"""Dfg builder that allows tracking a set of wires and appending operations by index."""

from collections.abc import Iterable
from typing import Any

from hugr import tys
from hugr.build.dfg import Dfg
from hugr.hugr.node_port import Node, Wire
from hugr.ops import Command, ComWire


class TrackedDfg(Dfg):
    """Dfg builder to append operations to wires by index.

    Args:
        *input_types: Input types of the Dfg.
        track_inputs: Whether to track the input wires.

    Examples:
        >>> dfg = TrackedDfg(tys.Bool, tys.Unit, track_inputs=True)
        >>> dfg.tracked
        [OutPort(Node(1), 0), OutPort(Node(1), 1)]
    """

    #: Tracked wires. None if index is no longer tracked.
    tracked: list[Wire | None]

    def __init__(self, *input_types: tys.Type, track_inputs: bool = False) -> None:
        super().__init__(*input_types)
        self.tracked = list(self.inputs()) if track_inputs else []

    def track_wire(self, wire: Wire) -> int:
        """Add a wire from this DFG to the tracked wires, and return its index.

        Args:
            wire: Wire to track.

        Returns:
            Index of the tracked wire.

        Examples:
            >>> dfg = TrackedDfg(tys.Bool, tys.Unit)
            >>> dfg.track_wire(dfg.inputs()[0])
            0
        """
        self.tracked.append(wire)
        return len(self.tracked) - 1

    def untrack_wire(self, index: int) -> Wire:
        """Untrack a wire by index and return it.

        Args:
            index: Index of the wire to untrack.

        Returns:
            Wire that was untracked.

        Raises:
            IndexError: If the index is not a tracked wire.

        Examples:
            >>> dfg = TrackedDfg(tys.Bool, tys.Unit)
            >>> w = dfg.inputs()[0]
            >>> idx = dfg.track_wire(w)
            >>> dfg.untrack_wire(idx) == w
            True
        """
        w = self.tracked_wire(index)
        self.tracked[index] = None
        return w

    def track_wires(self, wires: Iterable[Wire]) -> list[int]:
        """Set a list of wires to be tracked and return their indices.

        Args:
            wires: Wires to track.

        Returns:
            List of indices of the tracked wires.

        Examples:
            >>> dfg = TrackedDfg(tys.Bool, tys.Unit)
            >>> dfg.track_wires(dfg.inputs())
            [0, 1]
        """
        return [self.track_wire(w) for w in wires]

    def track_inputs(self) -> list[int]:
        """Track all input wires and return their indices.

        Returns:
            List of indices of the tracked input wires.

        Examples:
            >>> dfg = TrackedDfg(tys.Bool, tys.Unit)
            >>> dfg.track_inputs()
            [0, 1]
        """
        return self.track_wires(self.inputs())

    def tracked_wire(self, index: int) -> Wire:
        """Get the tracked wire at the given index.

        Args:
            index: Index of the tracked wire.

        Raises:
            IndexError: If the index is not a tracked wire.

        Returns:
            Tracked wire

        Examples:
            >>> dfg = TrackedDfg(tys.Bool, tys.Unit, track_inputs=True)
            >>> dfg.tracked_wire(0) == dfg.inputs()[0]
            True
        """
        try:
            tracked = self.tracked[index]
        except IndexError:
            tracked = None
        if tracked is None:
            msg = f"Index {index} not a tracked wire."
            raise IndexError(msg)
        return tracked

    def add(self, com: Command, *, metadata: dict[str, Any] | None = None) -> Node:
        """Add a command to the DFG.

        Overrides :meth:`Dfg.add <hugr.dfg.Dfg.add>` to allow Command inputs
        to be either :class:`Wire <hugr.node_port.Wire>` or indices to tracked wires.

        Any incoming :class:`Wire <hugr.node_port.Wire>` will
        be connected directly, while any integer will be treated as a reference
        to the tracked wire at that index.

        Any tracked wires will be updated to the output of the new node at the same port
        as the incoming index.

        Args:
            com: Command to append.
            metadata: Metadata to attach to the function definition. Defaults to None.

        Returns:
            The new node.

        Raises:
            IndexError: If any input index is not a tracked wire.

        Examples:
            >>> dfg = TrackedDfg(tys.Bool, track_inputs=True)
            >>> dfg.tracked
            [OutPort(Node(1), 0)]
            >>> dfg.add(ops.Noop()(0))
            Node(3)
            >>> dfg.tracked
            [OutPort(Node(3), 0)]
        """
        wires = self._to_wires(com.incoming)
        n = self.add_op(com.op, *wires)

        for port_offset, com_wire in enumerate(com.incoming):
            if isinstance(com_wire, int):
                tracked_idx = com_wire
            else:
                continue
            # update tracked wires to matching port outputs of new node
            self.tracked[tracked_idx] = n.out(port_offset)

        return n

    def _to_wires(self, in_wires: Iterable[ComWire]) -> Iterable[Wire]:
        return (
            self.tracked_wire(inc) if isinstance(inc, int) else inc for inc in in_wires
        )

    def set_indexed_outputs(self, *in_wires: ComWire) -> None:
        """Set the Dfg outputs, using either :class:`Wire <hugr.node_port.Wire>` or
        indices to tracked wires.

        Args:
            *in_wires: Wires/indices to set as outputs.

        Raises:
            IndexError: If any input index is not a tracked wire.

        Examples:
            >>> dfg = TrackedDfg(tys.Bool, tys.Unit)
            >>> (b, i) = dfg.inputs()
            >>> dfg.track_wire(b)
            0
            >>> dfg.set_indexed_outputs(0, i)
        """
        self.set_outputs(*self._to_wires(in_wires))

    def set_tracked_outputs(self) -> None:
        """Set the Dfg outputs to the tracked wires.


        Examples:
            >>> dfg = TrackedDfg(tys.Bool, tys.Unit, track_inputs=True)
            >>> dfg.set_tracked_outputs()
        """
        self.set_outputs(*(w for w in self.tracked if w is not None))



"""HUGR extensions."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, TypeVar

from semver import Version

import hugr._serialization.extension as ext_s
from hugr import ops, tys
from hugr.utils import ser_it

__all__ = [
    "ExplicitBound",
    "FromParamsBound",
    "TypeDef",
    "FixedHugr",
    "OpDefSig",
    "OpDef",
    "Extension",
    "Version",
]

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from hugr.hugr import Hugr
    from hugr.tys import ExtensionId


@dataclass
class ExplicitBound:
    """An explicit type bound on an :class:`OpDef`.


    Examples:
        >>> ExplicitBound(tys.TypeBound.Copyable)
        ExplicitBound(bound=<TypeBound.Copyable: 'C'>)
    """

    bound: tys.TypeBound

    def _to_serial(self) -> ext_s.ExplicitBound:
        return ext_s.ExplicitBound(bound=self.bound)

    def _to_serial_root(self) -> ext_s.TypeDefBound:
        return ext_s.TypeDefBound(root=self._to_serial())


@dataclass
class FromParamsBound:
    """Calculate the type bound of an :class:`OpDef` from the join of its parameters at
    the given indices.


    Examples:
        >>> FromParamsBound(indices=[0, 1])
        FromParamsBound(indices=[0, 1])
    """

    indices: list[int]

    def _to_serial(self) -> ext_s.FromParamsBound:
        return ext_s.FromParamsBound(indices=self.indices)

    def _to_serial_root(self) -> ext_s.TypeDefBound:
        return ext_s.TypeDefBound(root=self._to_serial())


@dataclass
class NoParentExtension(Exception):
    """Parent extension must be set."""

    kind: str

    def __str__(self):
        return f"{self.kind} does not belong to an extension."


@dataclass(init=False)
class ExtensionObject:
    """An object associated with an :class:`Extension`."""

    _extension: Extension | None = field(
        default=None, init=False, repr=False, compare=False
    )

    def get_extension(self) -> Extension:
        """Retrieve the extension associated with the object.

        Returns:
            The extension associated with the object.

        Raises:
            NoParentExtension: If the object is not associated with an extension.
        """
        if self._extension is None:
            msg = self.__class__.__name__
            raise NoParentExtension(msg)
        return self._extension


@dataclass
class TypeDef(ExtensionObject):
    """Type definition in an :class:`Extension`.


    Examples:
        >>> td = TypeDef(
        ...     name="MyType",
        ...     description="A type definition.",
        ...     params=[tys.TypeTypeParam(tys.TypeBound.Copyable)],
        ...     bound=FromParamsBound([0]),
        ... )
        >>> td.name
        'MyType'
    """

    #: The name of the type.
    name: str
    #: A description of the type.
    description: str
    #: The type parameters of the type if polymorphic.
    params: list[tys.TypeParam]
    #: The type bound of the type.
    bound: ExplicitBound | FromParamsBound

    def _to_serial(self) -> ext_s.TypeDef:
        return ext_s.TypeDef(
            extension=self.get_extension().name,
            name=self.name,
            description=self.description,
            params=ser_it(self.params),
            bound=ext_s.TypeDefBound(root=self.bound._to_serial()),
        )

    def instantiate(self, args: Sequence[tys.TypeArg]) -> tys.ExtType:
        """Instantiate a concrete type from this type definition.

        Args:
            args: Type arguments corresponding to the type parameters of the definition.
        """
        return tys.ExtType(self, list(args))


@dataclass
class FixedHugr:
    """A HUGR used to define lowerings of operations in an :class:`OpDef`."""

    #: Extensions used in the HUGR.
    extensions: tys.ExtensionSet
    #: HUGR defining operation lowering.
    hugr: Hugr

    def _to_serial(self) -> ext_s.FixedHugr:
        return ext_s.FixedHugr(extensions=self.extensions, hugr=self.hugr.to_str())


@dataclass
class OpDefSig:
    """Type signature of an :class:`OpDef`.

    Args:
        poly_func: The polymorphic function type of the operation.
        binary: If no static type scheme known, flag indicates a computation of the
            signature
    """

    #: The polymorphic function type of the operation (type scheme).
    poly_func: tys.PolyFuncType | None
    #: If no static type scheme known, flag indicates a computation of the signature.
    binary: bool

    def __init__(
        self,
        poly_func: tys.PolyFuncType | tys.FunctionType | None,
        binary: bool = False,
    ) -> None:
        if poly_func is None and not binary:
            msg = (
                "Signature must be provided if binary"
                " signature computation is not expected."
            )
            raise ValueError(msg)
        if isinstance(poly_func, tys.FunctionType):
            poly_func = tys.PolyFuncType([], poly_func)
        self.poly_func = poly_func
        self.binary = binary


@dataclass
class OpDef(ExtensionObject):
    """Operation definition in an :class:`Extension`."""

    #: The name of the operation.
    name: str
    #: The type signature of the operation.
    signature: OpDefSig
    #: A description of the operation.
    description: str = ""
    #: Miscellaneous information about the operation.
    misc: dict[str, Any] = field(default_factory=dict)
    #: Lowerings of the operation.
    lower_funcs: list[FixedHugr] = field(default_factory=list, repr=False)

    def _to_serial(self) -> ext_s.OpDef:
        return ext_s.OpDef(
            extension=self.get_extension().name,
            name=self.name,
            description=self.description,
            misc=self.misc,
            signature=self.signature.poly_func._to_serial()
            if self.signature.poly_func
            else None,
            binary=self.signature.binary,
            lower_funcs=[f._to_serial() for f in self.lower_funcs],
        )

    def qualified_name(self) -> str:
        ext_name = self._extension.name if self._extension else ""
        if ext_name:
            return f"{ext_name}.{self.name}"
        return self.name

    def instantiate(
        self,
        args: Sequence[tys.TypeArg] | None = None,
        concrete_signature: tys.FunctionType | None = None,
    ) -> ops.ExtOp:
        """Instantiate an operation from this definition.

        Args:
            args: Type arguments corresponding to the type parameters of the definition.
            concrete_signature: Concrete function type of the operation, only required
            if the operation is polymorphic.
        """
        return ops.ExtOp(self, concrete_signature, list(args or []))


T = TypeVar("T", bound=ops.RegisteredOp)


@dataclass
class Extension:
    """HUGR extension declaration."""

    #: The name of the extension.
    name: ExtensionId
    #: The version of the extension.
    version: Version
    #: Type definitions in the extension.
    types: dict[str, TypeDef] = field(default_factory=dict)
    #: Operation definitions in the extension.
    operations: dict[str, OpDef] = field(default_factory=dict)

    @dataclass
    class NotFound(Exception):
        """An object was not found in the extension."""

        name: str

    def _to_serial(self) -> ext_s.Extension:
        return ext_s.Extension(
            name=self.name,
            version=self.version,  # type: ignore[arg-type]
            types={k: v._to_serial() for k, v in self.types.items()},
            operations={k: v._to_serial() for k, v in self.operations.items()},
        )

    def to_json(self) -> str:
        """Serialize the extension to a JSON string."""
        return self._to_serial().model_dump_json()

    @classmethod
    def from_json(cls, json_str: str) -> Extension:
        """Deserialize a JSON string to a Extension object.

        Args:
            json_str: The JSON string representing a Extension.

        Returns:
            The deserialized Extension object.
        """
        return ext_s.Extension.model_validate_json(json_str).deserialize()

    def add_op_def(self, op_def: OpDef) -> OpDef:
        """Add an operation definition to the extension.

        Args:
            op_def: The operation definition to add.

        Returns:
            The added operation definition, now associated with the extension.
        """
        op_def._extension = self
        self.operations[op_def.name] = op_def
        return self.operations[op_def.name]

    def add_type_def(self, type_def: TypeDef) -> TypeDef:
        """Add a type definition to the extension.

        Args:
            type_def: The type definition to add.

        Returns:
            The added type definition, now associated with the extension.
        """
        type_def._extension = self
        self.types[type_def.name] = type_def
        return self.types[type_def.name]

    @dataclass
    class OperationNotFound(NotFound):
        """Operation not found in extension."""

    def get_op(self, name: str) -> OpDef:
        """Retrieve an operation definition by name.

        Args:
            name: The name of the operation.

        Returns:
            The operation definition.

        Raises:
            OperationNotFound: If the operation is not found in the extension.
        """
        try:
            return self.operations[name]
        except KeyError as e:
            raise self.OperationNotFound(name) from e

    @dataclass
    class TypeNotFound(NotFound):
        """Type not found in extension."""

    def get_type(self, name: str) -> TypeDef:
        """Retrieve a type definition by name.

        Args:
            name: The name of the type.

        Returns:
            The type definition.

        Raises:
            TypeNotFound: If the type is not found in the extension.
        """
        try:
            return self.types[name]
        except KeyError as e:
            raise self.TypeNotFound(name) from e

    @dataclass
    class ValueNotFound(NotFound):
        """Value not found in extension."""

    T = TypeVar("T", bound=ops.RegisteredOp)

    def register_op(
        self,
        name: str | None = None,
        signature: OpDefSig | tys.PolyFuncType | tys.FunctionType | None = None,
        description: str | None = None,
        misc: dict[str, Any] | None = None,
        lower_funcs: list[FixedHugr] | None = None,
    ) -> Callable[[type[T]], type[T]]:
        """Register a class as corresponding to an operation definition.

        If `name` is not provided, the class name is used.
        If `signature` is not provided, a binary signature is assumed.
        If `description` is not provided, the class docstring is used.

        See :class:`OpDef` for other parameters.
        """
        if not isinstance(signature, OpDefSig):
            binary = signature is None
            signature = OpDefSig(signature, binary)

        def _inner(cls: type[T]) -> type[T]:
            new_description = cls.__doc__ if description is None and cls.__doc__ else ""
            new_name = cls.__name__ if name is None else name
            op_def = self.add_op_def(
                OpDef(
                    new_name,
                    signature,
                    new_description,
                    misc or {},
                    lower_funcs or [],
                )
            )
            cls.const_op_def = op_def
            return cls

        return _inner


@dataclass
class ExtensionRegistry:
    """Registry of extensions."""

    #: Extensions in the registry, indexed by name.
    extensions: dict[ExtensionId, Extension] = field(default_factory=dict)

    @dataclass
    class ExtensionNotFound(Exception):
        """Extension not found in registry."""

        extension_id: ExtensionId

    @dataclass
    class ExtensionExists(Exception):
        """Extension already exists in registry."""

        extension_id: ExtensionId

    def add_extension(self, extension: Extension) -> Extension:
        """Add an extension to the registry.

        Args:
            extension: The extension to add.

        Returns:
            The added extension.

        Raises:
            ExtensionExists: If an extension with the same name already exists.
        """
        if extension.name in self.extensions:
            raise self.ExtensionExists(extension.name)
        self.extensions[extension.name] = extension
        return self.extensions[extension.name]

    def get_extension(self, name: ExtensionId) -> Extension:
        """Retrieve an extension by name.

        Args:
            name: The name of the extension.

        Returns:
            Extension in the registry.

        Raises:
            ExtensionNotFound: If the extension is not found in the registry.
        """
        try:
            return self.extensions[name]
        except KeyError as e:
            raise self.ExtensionNotFound(name) from e



"""HUGR values, used for static constants in HUGR programs."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Protocol, cast, runtime_checkable

import hugr._serialization.ops as sops
import hugr._serialization.tys as stys
import hugr.model as model
from hugr import tys
from hugr.utils import comma_sep_repr, comma_sep_str, ser_it

if TYPE_CHECKING:
    from collections.abc import Iterable

    from hugr.hugr import Hugr


@runtime_checkable
class Value(Protocol):
    """Abstract value definition. Must be serializable into a HUGR value."""

    def _to_serial(self) -> sops.BaseValue:
        """Convert to serializable model."""
        ...  # pragma: no cover

    def _to_serial_root(self) -> sops.Value:
        return sops.Value(root=self._to_serial())  # type: ignore[arg-type]

    def type_(self) -> tys.Type:
        """Report the type of the value.

        Example:
            >>> TRUE.type_()
            Bool
        """
        ...  # pragma: no cover

    def to_model(self) -> model.Term: ...


@dataclass
class Sum(Value):
    """Sum-of-product value.

    Example:
        >>> Sum(0, tys.Sum([[tys.Bool], [tys.Unit]]), [TRUE])
        Sum(tag=0, typ=Sum([[Bool], [Unit]]), vals=[TRUE])
    """

    #: Tag identifying the variant.
    tag: int
    #: Type of the sum: defines all possible variants.
    typ: tys.Sum
    #: The values of this variant row.
    vals: list[Value]

    @property
    def n_variants(self) -> int:
        return len(self.typ.variant_rows)

    def type_(self) -> tys.Sum:
        return self.typ

    def _to_serial(self) -> sops.SumValue:
        return sops.SumValue(
            tag=self.tag,
            typ=stys.SumType(root=self.type_()._to_serial()),
            vs=ser_it(self.vals),
        )

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Sum)
            and self.tag == other.tag
            and self.typ == other.typ
            and self.vals == other.vals
        )

    def to_model(self) -> model.Term:
        variants = [
            model.List([type.to_model() for type in row])
            for row in self.typ.variant_rows
        ]
        types = [
            model.Apply("core.const", [cast(model.Term, type)])
            for type in variants[self.tag].parts
        ]
        values = [value.to_model() for value in self.vals]

        return model.Apply(
            "core.const.adt",
            [
                model.List(variants),
                model.List(types),
                model.Literal(self.tag),
                model.Tuple(values),
            ],
        )


class UnitSum(Sum):
    """Simple :class:`Sum` with each variant being an empty row.

    Example:
        >>> UnitSum(0, 3)
        UnitSum(0, 3)
        >>> UnitSum(0, 1)
        Unit
        >>> assert UnitSum(0, 2) == FALSE
        >>> assert UnitSum(1, 2) == TRUE
    """

    def __init__(self, tag: int, size: int):
        super().__init__(
            tag=tag,
            typ=tys.UnitSum(size),
            vals=[],
        )

    def __repr__(self) -> str:
        if self == TRUE:
            return "TRUE"
        if self == FALSE:
            return "FALSE"
        if self == Unit:
            return "Unit"
        return f"UnitSum({self.tag}, {self.n_variants})"


def bool_value(b: bool) -> UnitSum:
    """Convert a python bool to a HUGR boolean value.

    Example:
        >>> bool_value(True)
        TRUE
        >>> bool_value(False)
        FALSE
    """
    return UnitSum(int(b), 2)


#: HUGR unit type. Sum with a single empty row variant.
Unit = UnitSum(0, 1)
#: HUGR true value.
TRUE = bool_value(True)
#: HUGR false value.
FALSE = bool_value(False)


@dataclass(eq=False)
class Tuple(Sum):
    """Tuple or product value, defined by a list of values.
    Internally a :class:`Sum` with a single variant row.

    Example:
        >>> tup = Tuple(TRUE, FALSE)
        >>> tup
        Tuple(TRUE, FALSE)
        >>> tup.type_()
        Tuple(Bool, Bool)

    """

    def __init__(self, *vals: Value):
        val_list = list(vals)
        super().__init__(
            tag=0, typ=tys.Tuple(*(v.type_() for v in val_list)), vals=val_list
        )

    # sops.TupleValue isn't an instance of sops.SumValue
    # so mypy doesn't like the override of Sum._to_serial
    def _to_serial(self) -> sops.TupleValue:  # type: ignore[override]
        return sops.TupleValue(
            vs=ser_it(self.vals),
        )

    def __repr__(self) -> str:
        return f"Tuple({comma_sep_repr(self.vals)})"


@dataclass(eq=False)
class Some(Sum):
    """Optional tuple of value, containing a list of values.

    Example:
        >>> some = Some(TRUE, FALSE)
        >>> str(some)
        'Some(TRUE, FALSE)'
        >>> some.type_()
        Option(Bool, Bool)

    """

    def __init__(self, *vals: Value):
        val_list = list(vals)
        super().__init__(
            tag=1, typ=tys.Option(*(v.type_() for v in val_list)), vals=val_list
        )

    def __repr__(self) -> str:
        return f"Some({comma_sep_repr(self.vals)})"


@dataclass(eq=False)
class None_(Sum):
    """Optional tuple of value, containing no values.

    Example:
        >>> none = None_(tys.Bool)
        >>> str(none)
        'None'
        >>> none.type_()
        Option(Bool)

    """

    def __init__(self, *types: tys.Type):
        super().__init__(tag=0, typ=tys.Option(*types), vals=[])

    def __repr__(self) -> str:
        return f"None({comma_sep_str(self.typ.variant_rows[1])})"

    def __str__(self) -> str:
        return "None"


@dataclass(eq=False)
class Left(Sum):
    """Left variant of a :class:`tys.Either` type, containing a list of values.

    In fallible contexts, this represents the failure variant.

    Example:
        >>> left = Left([TRUE, FALSE], [tys.Bool])
        >>> str(left)
        'Left(TRUE, FALSE)'
        >>> str(left.type_())
        'Either((Bool, Bool), Bool)'
    """

    def __init__(self, vals: Iterable[Value], right_typ: Iterable[tys.Type]):
        val_list = list(vals)
        super().__init__(
            tag=0,
            typ=tys.Either([v.type_() for v in val_list], right_typ),
            vals=val_list,
        )

    def __repr__(self) -> str:
        _, right_typ = self.typ.variant_rows
        return f"Left(vals={self.vals}, right_typ={list(right_typ)})"

    def __str__(self) -> str:
        return f"Left({comma_sep_str(self.vals)})"


@dataclass(eq=False)
class Right(Sum):
    """Right variant of a :class:`tys.Either` type, containing a list of values.

    In fallible contexts, this represents the success variant.

    Internally a :class:`Sum` with two variant rows.

    Example:
        >>> right = Right([tys.Bool, tys.Bool, tys.Bool], [TRUE, FALSE])
        >>> str(right)
        'Right(TRUE, FALSE)'
        >>> str(right.type_())
        'Either((Bool, Bool, Bool), (Bool, Bool))'
    """

    def __init__(self, left_typ: Iterable[tys.Type], vals: Iterable[Value]):
        val_list = list(vals)
        super().__init__(
            tag=1,
            typ=tys.Either(left_typ, [v.type_() for v in val_list]),
            vals=val_list,
        )

    def __repr__(self) -> str:
        left_typ, _ = self.typ.variant_rows
        return f"Right(left_typ={list(left_typ)}, vals={self.vals})"

    def __str__(self) -> str:
        return f"Right({comma_sep_str(self.vals)})"


@dataclass
class Function(Value):
    """Higher order function value, defined by a :class:`Hugr <hugr.hugr.HUGR>`."""

    body: Hugr

    def type_(self) -> tys.FunctionType:
        return self.body.entrypoint_op().inner_signature()

    def _to_serial(self) -> sops.FunctionValue:
        return sops.FunctionValue(
            hugr=self.body._to_serial(),
        )

    def to_model(self) -> model.Term:
        module = self.body.to_model()
        return model.Func(module.root)


@dataclass
class Extension(Value):
    """Non-core extension value."""

    #: Value name.
    name: str
    #: Value type.
    typ: tys.Type
    #: Value payload.
    val: Any

    #: Extension set. Deprecated, no longer used. Will be removed in hugr v0.13.
    extensions: tys.ExtensionSet = field(default_factory=tys.ExtensionSet)

    def type_(self) -> tys.Type:
        return self.typ

    def _to_serial(self) -> sops.CustomValue:
        return sops.CustomValue(
            typ=self.typ._to_serial_root(),
            value=sops.CustomConst(c=self.name, v=self.val),
        )

    def to_model(self) -> model.Term:
        type = cast(model.Term, self.typ.to_model())
        json = sops.CustomConst(c=self.name, v=self.val).model_dump_json()
        return model.Apply("compat.const_json", [type, model.Literal(json)])


class ExtensionValue(Value, Protocol):
    """Protocol which types can implement to be a HUGR extension value."""

    def to_value(self) -> Extension:
        """Convert to a HUGR extension value."""
        ...  # pragma: no cover

    def type_(self) -> tys.Type:
        return self.to_value().type_()

    def _to_serial(self) -> sops.CustomValue:
        return self.to_value()._to_serial()

    def to_model(self) -> model.Term:
        # Fallback
        return self.to_value().to_model()



from pydantic import ConfigDict

from .extension import OpDef
from .ops import OpType, Value
from .ops import classes as ops_classes
from .serial_hugr import VersionField
from .tys import ConfiguredBaseModel, PolyFuncType, SumType, Type, model_rebuild


class TestingHugr(ConfiguredBaseModel):
    """A serializable representation of a Hugr Type, SumType, PolyFuncType,
    Value, OpType. Intended for testing only.
    """

    version: str = VersionField
    typ: Type | None = None
    sum_type: SumType | None = None
    poly_func_type: PolyFuncType | None = None
    value: Value | None = None
    optype: OpType | None = None
    op_def: OpDef | None = None

    @classmethod
    def get_version(cls) -> str:
        """Return the version of the schema."""
        return cls().version

    @classmethod
    def _pydantic_rebuild(cls, config: ConfigDict | None = None, **kwargs):
        config = config or ConfigDict()
        my_classes = dict(ops_classes)
        my_classes[cls.__name__] = cls
        model_rebuild(my_classes, config=config, **kwargs)



import pytest

from hugr.hugr.node_port import InPort, Node, OutPort
from hugr.ops import (
    CFG,
    DFG,
    AliasDecl,
    AliasDefn,
    Call,
    CallIndirect,
    Case,
    Conditional,
    Const,
    DataflowBlock,
    ExitBlock,
    FuncDecl,
    FuncDefn,
    Input,
    LoadConst,
    LoadFunc,
    MakeTuple,
    Module,
    Noop,
    Op,
    Output,
    Tag,
    TailLoop,
    UnpackTuple,
)
from hugr.std.int import DivMod
from hugr.std.logic import Not
from hugr.tys import Bool, OrderKind, PolyFuncType, TypeBound
from hugr.val import TRUE


@pytest.mark.parametrize(
    ("op", "string"),
    [
        (Input([]), "Input"),
        (Output([]), "Output"),
        (Not, "logic.Not"),
        (DivMod, "arithmetic.int.idivmod_u<5>"),
        (MakeTuple(), "MakeTuple"),
        (UnpackTuple(), "UnpackTuple"),
        (Tag(0, Bool), "Tag(0)"),
        (CFG([]), "CFG"),
        (DFG([]), "DFG"),
        (DataflowBlock([]), "DataflowBlock"),
        (ExitBlock([]), "ExitBlock"),
        (LoadConst(), "LoadConst"),
        (Conditional(Bool, []), "Conditional"),
        (TailLoop([], []), "TailLoop"),
        (Case([]), "Case"),
        (Module(), "Module"),
        (Call(PolyFuncType.empty()), "Call"),
        (CallIndirect(), "CallIndirect"),
        (LoadFunc(PolyFuncType.empty()), "LoadFunc"),
        (FuncDefn("foo", []), "FuncDefn(foo)"),
        (FuncDecl("bar", PolyFuncType.empty()), "FuncDecl(bar)"),
        (Const(TRUE), "Const(TRUE)"),
        (Noop(), "Noop"),
        (AliasDecl("baz", TypeBound.Any), "AliasDecl(baz)"),
        (AliasDefn("baz", Bool), "AliasDefn(baz)"),
    ],
)
def test_ops_str(op: Op, string: str):
    assert op.name() == string


@pytest.mark.parametrize(
    "op",
    [Noop(), Call(PolyFuncType.empty()), LoadFunc(PolyFuncType.empty()), LoadConst()],
)
def test_order_ports(op: Op):
    assert op.port_kind(InPort(Node(0), -1)) == OrderKind()
    assert op.port_kind(OutPort(Node(0), -1)) == OrderKind()



"""Helpers to export hugr graphs from their python representation to hugr model."""

import json
from collections.abc import Sequence
from typing import Generic, TypeVar, cast

import hugr.model as model
from hugr.hugr.base import Hugr, Node
from hugr.hugr.node_port import InPort, OutPort
from hugr.ops import (
    CFG,
    DFG,
    AliasDecl,
    AliasDefn,
    AsExtOp,
    Call,
    CallIndirect,
    Conditional,
    Const,
    Custom,
    DataflowBlock,
    ExitBlock,
    FuncDecl,
    FuncDefn,
    Input,
    LoadConst,
    LoadFunc,
    Output,
    Tag,
    TailLoop,
)
from hugr.tys import ConstKind, FunctionKind, Type, TypeBound, TypeParam, TypeTypeParam


class ModelExport:
    """Helper to export a Hugr."""

    def __init__(self, hugr: Hugr):
        self.hugr = hugr
        self.link_ports: _UnionFind[InPort | OutPort] = _UnionFind()
        self.link_names: dict[InPort | OutPort, str] = {}

        # TODO: Store the hugr entrypoint

        for a, b in self.hugr.links():
            self.link_ports.union(a, b)

    def link_name(self, port: InPort | OutPort) -> str:
        """Return the name of the link that a given port is connected to."""
        root = self.link_ports[port]

        if root in self.link_names:
            return self.link_names[root]
        else:
            index = str(len(self.link_names))
            self.link_names[root] = index
            return index

    def export_node(self, node: Node) -> model.Node | None:
        """Export the node with the given node id."""
        node_data = self.hugr[node]

        inputs = [self.link_name(InPort(node, i)) for i in range(node_data._num_inps)]
        outputs = [self.link_name(OutPort(node, i)) for i in range(node_data._num_outs)]
        meta = []

        # Export JSON metadata
        for meta_name, meta_value in node_data.metadata.items():
            # TODO: Is this the correct way to convert the metadata as JSON?
            meta_json = json.dumps(meta_value)
            meta.append(
                model.Apply(
                    "compat.meta_json",
                    [model.Literal(meta_name), model.Literal(meta_json)],
                )
            )

        # Add an order hint key to the node if necessary
        if _needs_order_key(self.hugr, node):
            meta.append(model.Apply("core.order_hint.key", [model.Literal(node.idx)]))

        match node_data.op:
            case DFG() as op:
                region = self.export_region_dfg(node)

                return model.Node(
                    operation=model.Dfg(),
                    regions=[region],
                    signature=op.outer_signature().to_model(),
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case Custom() as op:
                name = f"{op.extension}.{op.op_name}"
                args = cast(list[model.Term], [arg.to_model() for arg in op.args])
                signature = op.signature.to_model()

                return model.Node(
                    operation=model.CustomOp(model.Apply(name, args)),
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case AsExtOp() as op:
                name = op.op_def().qualified_name()
                args = cast(
                    list[model.Term], [arg.to_model() for arg in op.type_args()]
                )
                signature = op.outer_signature().to_model()

                return model.Node(
                    operation=model.CustomOp(model.Apply(name, args)),
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case Conditional() as op:
                regions = [
                    self.export_region_dfg(child) for child in node_data.children
                ]

                signature = op.outer_signature().to_model()

                return model.Node(
                    operation=model.Conditional(),
                    regions=regions,
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case TailLoop() as op:
                region = self.export_region_dfg(node)
                signature = op.outer_signature().to_model()
                return model.Node(
                    operation=model.TailLoop(),
                    regions=[region],
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case FuncDefn() as op:
                name = _mangle_name(node, op.f_name)
                symbol = self.export_symbol(
                    name, op.signature.params, op.signature.body
                )
                region = self.export_region_dfg(node)

                return model.Node(
                    operation=model.DefineFunc(symbol), regions=[region], meta=meta
                )

            case FuncDecl() as op:
                name = _mangle_name(node, op.f_name)
                symbol = self.export_symbol(
                    name, op.signature.params, op.signature.body
                )
                return model.Node(operation=model.DeclareFunc(symbol), meta=meta)

            case AliasDecl() as op:
                symbol = model.Symbol(name=op.alias, signature=model.Apply("core.type"))

                return model.Node(operation=model.DeclareAlias(symbol), meta=meta)

            case AliasDefn() as op:
                symbol = model.Symbol(name=op.alias, signature=model.Apply("core.type"))

                alias_value = cast(model.Term, op.definition.to_model())

                return model.Node(
                    operation=model.DefineAlias(symbol, alias_value), meta=meta
                )

            case Call() as op:
                input_types = [type.to_model() for type in op.instantiation.input]
                output_types = [type.to_model() for type in op.instantiation.output]
                signature = op.instantiation.to_model()
                func_args = cast(
                    list[model.Term], [type.to_model() for type in op.type_args]
                )
                func_name = self.find_func_input(node)

                if func_name is None:
                    error = f"Call node {node} is not connected to a function."
                    raise ValueError(error)

                func = model.Apply(func_name, func_args)

                return model.Node(
                    operation=model.CustomOp(
                        model.Apply(
                            "core.call",
                            [
                                model.List(input_types),
                                model.List(output_types),
                                func,
                            ],
                        )
                    ),
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case LoadFunc() as op:
                signature = op.instantiation.to_model()
                func_args = cast(
                    list[model.Term], [type.to_model() for type in op.type_args]
                )
                func_name = self.find_func_input(node)

                if func_name is None:
                    error = f"LoadFunc node {node} is not connected to a function."
                    raise ValueError(error)

                func = model.Apply(func_name, func_args)

                return model.Node(
                    operation=model.CustomOp(
                        model.Apply("core.load_const", [signature, func])
                    ),
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case CallIndirect() as op:
                input_types = [type.to_model() for type in op.signature.input]
                output_types = [type.to_model() for type in op.signature.output]

                func = model.Apply(
                    "core.fn",
                    [model.List(input_types), model.List(output_types)],
                )

                signature = model.Apply(
                    "core.fn",
                    [
                        model.List([func, *input_types]),
                        model.List(output_types),
                    ],
                )

                return model.Node(
                    operation=model.CustomOp(
                        model.Apply(
                            "core.call_indirect",
                            [
                                model.List(input_types),
                                model.List(output_types),
                            ],
                        )
                    ),
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case LoadConst() as op:
                value = self.find_const_input(node)

                if value is None:
                    error = f"LoadConst node {node} is not connected to a constant."
                    raise ValueError(error)

                type = cast(model.Term, op.type_.to_model())
                signature = op.outer_signature().to_model()

                return model.Node(
                    operation=model.CustomOp(
                        model.Apply("core.load_const", [type, value])
                    ),
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    meta=meta,
                )

            case Const() as op:
                return None

            case CFG() as op:
                signature = op.outer_signature().to_model()
                region = self.export_region_cfg(node)

                return model.Node(
                    operation=model.Cfg(),
                    signature=signature,
                    inputs=inputs,
                    outputs=outputs,
                    regions=[region],
                    meta=meta,
                )

            case DataflowBlock() as op:
                region = self.export_region_dfg(node)

                input_types = [
                    model.Apply(
                        "core.ctrl",
                        [model.List([type.to_model() for type in op.inputs])],
                    )
                ]

                other_output_types = [type.to_model() for type in op.other_outputs]
                output_types = [
                    model.Apply(
                        "core.ctrl",
                        [
                            model.List(
                                [
                                    *[type.to_model() for type in row],
                                    *other_output_types,
                                ]
                            )
                        ],
                    )
                    for row in op.sum_ty.variant_rows
                ]

                signature = model.Apply(
                    "core.fn",
                    [model.List(input_types), model.List(output_types)],
                )

                return model.Node(
                    operation=model.Block(),
                    inputs=inputs,
                    outputs=outputs,
                    regions=[region],
                    signature=signature,
                    meta=meta,
                )

            case Tag() as op:
                variants = model.List(
                    [
                        model.List([type.to_model() for type in row])
                        for row in op.sum_ty.variant_rows
                    ]
                )

                types = model.List(
                    [type.to_model() for type in op.sum_ty.variant_rows[op.tag]]
                )

                tag = model.Literal(op.tag)
                signature = op.outer_signature().to_model()

                return model.Node(
                    operation=model.CustomOp(
                        model.Apply("core.make_adt", [variants, types, tag])
                    ),
                    inputs=inputs,
                    outputs=outputs,
                    signature=signature,
                    meta=meta,
                )

            case op:
                error = f"Unknown operation: {op}"
                raise ValueError(error)

    def export_region_module(self, node: Node) -> model.Region:
        """Export a module node as a module region."""
        node_data = self.hugr[node]
        children = []

        for child in node_data.children:
            child_node = self.export_node(child)

            if child_node is not None:
                children.append(child_node)

        return model.Region(kind=model.RegionKind.MODULE, children=children)

    def export_region_dfg(self, node: Node) -> model.Region:
        """Export the children of a node as a dataflow region."""
        node_data = self.hugr[node]
        children: list[model.Node] = []
        source_types: model.Term = model.Wildcard()
        target_types: model.Term = model.Wildcard()
        sources = []
        targets = []
        meta = []

        for child in node_data.children:
            child_data = self.hugr[child]

            match child_data.op:
                case Input() as op:
                    source_types = model.List([type.to_model() for type in op.types])
                    sources = [
                        self.link_name(OutPort(child, i))
                        for i in range(child_data._num_outs)
                    ]

                case Output() as op:
                    target_types = model.List([type.to_model() for type in op.types])
                    targets = [
                        self.link_name(InPort(child, i))
                        for i in range(child_data._num_inps)
                    ]

                case _:
                    child_node = self.export_node(child)

                    if child_node is None:
                        continue

                    children.append(child_node)

                    meta += [
                        model.Apply(
                            "core.order_hint.order",
                            [model.Literal(child.idx), model.Literal(successor.idx)],
                        )
                        for successor in self.hugr.outgoing_order_links(child)
                        if not isinstance(self.hugr[successor].op, Output)
                    ]

        signature = model.Apply("core.fn", [source_types, target_types])

        return model.Region(
            kind=model.RegionKind.DATA_FLOW,
            signature=signature,
            children=children,
            sources=sources,
            targets=targets,
        )

    def export_region_cfg(self, node: Node) -> model.Region:
        """Export the children of a node as a control flow region."""
        node_data = self.hugr[node]

        source = None
        targets = []
        source_types: model.Term = model.Wildcard()
        target_types: model.Term = model.Wildcard()
        children = []

        for child in node_data.children:
            child_data = self.hugr[child]

            match child_data.op:
                case ExitBlock() as op:
                    target_types = model.List(
                        [type.to_model() for type in op.cfg_outputs]
                    )
                    targets = [
                        self.link_name(InPort(child, i))
                        for i in range(child_data._num_inps)
                    ]
                case DataflowBlock() as op:
                    if source is None:
                        source_types = model.List(
                            [type.to_model() for type in op.inputs]
                        )
                        source = self.link_name(OutPort(child, 0))

                    child_node = self.export_node(child)

                    if child_node is not None:
                        children.append(child_node)
                case _:
                    error = f"Unexpected operation in CFG {node}"
                    raise ValueError(error)

        if source is None:
            error = f"CFG {node} has no entry block."
            raise ValueError(error)

        signature = model.Apply("core.fn", [source_types, target_types])

        return model.Region(
            kind=model.RegionKind.CONTROL_FLOW,
            targets=targets,
            sources=[source],
            signature=signature,
            children=children,
        )

    def export_symbol(
        self, name: str, param_types: Sequence[TypeParam], body: Type
    ) -> model.Symbol:
        """Export a symbol."""
        constraints = []
        params = []

        for i, param_type in enumerate(param_types):
            param_name = str(i)

            params.append(model.Param(name=param_name, type=param_type.to_model()))

            match param_type:
                case TypeTypeParam(bound=TypeBound.Copyable):
                    constraints.append(
                        model.Apply("core.nonlinear", [model.Var(param_name)])
                    )
                case _:
                    pass

        return model.Symbol(
            name=name,
            params=params,
            constraints=constraints,
            signature=cast(model.Term, body.to_model()),
        )

    def find_func_input(self, node: Node) -> str | None:
        """Find the name of the function that a node is connected to, if any."""
        try:
            func_node = next(
                out_port.node
                for (in_port, out_ports) in self.hugr.incoming_links(node)
                if isinstance(self.hugr.port_kind(in_port), FunctionKind)
                for out_port in out_ports
            )
        except StopIteration:
            return None

        match self.hugr[func_node].op:
            case FuncDecl() as func_op:
                name = func_op.f_name
            case FuncDefn() as func_op:
                name = func_op.f_name
            case _:
                return None

        return _mangle_name(node, name)

    def find_const_input(self, node: Node) -> model.Term | None:
        """Find and export the constant that a node is connected to, if any."""
        try:
            const_node = next(
                out_port.node
                for (in_port, out_ports) in self.hugr.incoming_links(node)
                if isinstance(self.hugr.port_kind(in_port), ConstKind)
                for out_port in out_ports
            )
        except StopIteration:
            return None

        match self.hugr[const_node].op:
            case Const() as op:
                return op.val.to_model()
            case op:
                return None


def _mangle_name(node: Node, name: str) -> str:
    # Until we come to an agreement on the uniqueness of names, we mangle the names
    # by adding the node id.
    return f"_{name}_{node.idx}"


T = TypeVar("T")


class _UnionFind(Generic[T]):
    def __init__(self) -> None:
        self.parents: dict[T, T] = {}
        self.sizes: dict[T, int] = {}

    def __getitem__(self, item: T) -> T:
        if item not in self.parents:
            self.parents[item] = item
            self.sizes[item] = 1
            return item

        # Path splitting
        while self.parents[item] != item:
            parent = self.parents[item]
            self.parents[item] = self.parents[parent]
            item = parent

        return item

    def union(self, a: T, b: T):
        a = self[a]
        b = self[b]

        if a == b:
            return

        if self.sizes[a] < self.sizes[b]:
            (a, b) = (b, a)

        self.parents[b] = a
        self.sizes[a] += self.sizes[b]


def _needs_order_key(hugr: Hugr, node: Node) -> bool:
    """Checks whether the node has any order links for the purposes of
    exporting order hint metadata. Order links to `Input` or `Output`
    operations are ignored, since they are not present in the model format.
    """
    for succ in hugr.outgoing_order_links(node):
        succ_op = hugr[succ].op
        if not isinstance(succ_op, Output):
            return True

    for pred in hugr.incoming_order_links(node):
        pred_op = hugr[pred].op
        if not isinstance(pred_op, Input):
            return True

    return False



r"""Quantinuum system results.

Includes conversions to traditional distributions over bitstrings if a tagging
convention is used, including conversion to a pytket BackendResult.

Under this convention, tags are assumed to be a name of a bit register unless they fit
the regex pattern `^([a-z][\w_]*)\[(\d+)\]$` (like `my_Reg[12]`) in which case they
are assumed to refer to the nth element of a bit register.

For results of the form ``` result("<register>", value) ``` `value` can be `{0, 1}`,
wherein the register is assumed to be length 1, or lists over those values,
wherein the list is taken to be the value of the entire register.

For results of the form ``` result("<register>[n]", value) ``` `value` can only be
`{0,1}`.
The register is assumed to be at least `n+1` in size and unset
elements are assumed to be `0`.

Subsequent writes to the same register/element in the same shot will overwrite.

To convert to a `BackendResult` all registers must be present in all shots, and register
sizes cannot change between shots.

"""

from __future__ import annotations

import re
from collections import Counter, defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal

from typing_extensions import deprecated

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pytket.backends.backendresult import BackendResult


#: Primitive data types that can be returned by a result
DataPrimitive = int | float | bool
#: Data value that can be returned by a result: a primitive or a list of primitives
DataValue = DataPrimitive | list[DataPrimitive]
TaggedResult = tuple[str, DataValue]
# Pattern to match register index in tag, e.g. "reg[0]"
REG_INDEX_PATTERN = re.compile(r"^([a-z][\w_]*)\[(\d+)\]$")

BitChar = Literal["0", "1"]


@dataclass
class QsysShot:
    """Results from a single shot execution."""

    entries: list[TaggedResult] = field(default_factory=list)

    def __init__(self, entries: Iterable[TaggedResult] | None = None):
        self.entries = list(entries or [])

    def append(self, tag: str, data: DataValue) -> None:
        self.entries.append((tag, data))

    def as_dict(self) -> dict[str, DataValue]:
        """Convert results to a dictionary.

        For duplicate tags, the last value is used.

        Returns:
            dict: A dictionary where the keys are the tags and the
            values are the data.

        Example:
            >>> results = QsysShot()
            >>> results.append("tag1", 1)
            >>> results.append("tag2", 2)
            >>> results.append("tag2", 3)
            >>> results.as_dict()
            {'tag1': 1, 'tag2': 3}
        """
        return dict(self.entries)

    def to_register_bits(self) -> dict[str, str]:
        """Convert results to a dictionary of register bit values."""
        reg_bits: dict[str, list[BitChar]] = {}

        res_dict = self.as_dict()
        # relies on the fact that dict preserves insertion order
        for tag, data in res_dict.items():
            match = re.match(REG_INDEX_PATTERN, tag)
            if match is not None:
                reg_name, reg_index_str = match.groups()
                reg_index = int(reg_index_str)

                if reg_name not in reg_bits:
                    # Initialize register counts to False
                    reg_bits[reg_name] = ["0"] * (reg_index + 1)
                bitlst = reg_bits[reg_name]
                if reg_index >= len(bitlst):
                    # Extend register counts with "0"
                    bitlst += ["0"] * (reg_index - len(bitlst) + 1)

                bitlst[reg_index] = _cast_primitive_bit(data)
                continue
            match data:
                case list(vs):
                    reg_bits[tag] = [_cast_primitive_bit(v) for v in vs]
                case _:
                    reg_bits[tag] = [_cast_primitive_bit(data)]

        return {reg: "".join(bits) for reg, bits in reg_bits.items()}

    def collate_tags(self) -> dict[str, list[DataValue]]:
        """Collate all the entries with the same tag in to a dictionary with a list
        containing all the data for that tag.
        """
        tags: dict[str, list[DataValue]] = defaultdict(list)
        for tag, data in self.entries:
            tags[tag].append(data)
        return dict(tags)


@deprecated("Use QsysShot instead.")
class HResult(QsysShot):
    """Deprecated alias for QsysShot."""


def _cast_primitive_bit(data: DataValue) -> BitChar:
    if isinstance(data, int) and data in {0, 1}:
        return str(data)  # type: ignore[return-value]
    msg = f"Expected bit data for register value found {data}"
    raise ValueError(msg)


@dataclass
class QsysResult:
    """Results accumulated over multiple shots."""

    results: list[QsysShot]

    def __init__(
        self, results: Iterable[QsysShot | Iterable[TaggedResult]] | None = None
    ):
        self.results = [
            res if isinstance(res, QsysShot) else QsysShot(res) for res in results or []
        ]

    def register_counts(
        self, strict_names: bool = False, strict_lengths: bool = False
    ) -> dict[str, Counter[str]]:
        """Convert results to a dictionary of register counts.

        Returns:
            dict: A dictionary where the keys are the register names
            and the values are the counts of the register bitstrings.
        """
        return {
            reg: Counter(bitstrs)
            for reg, bitstrs in self.register_bitstrings(
                strict_lengths=strict_lengths, strict_names=strict_names
            ).items()
        }

    def register_bitstrings(
        self, strict_names: bool = False, strict_lengths: bool = False
    ) -> dict[str, list[str]]:
        """Convert results to a dictionary from register name to list of bitstrings over
        the shots.

        Args:
            strict_names: Whether to enforce that all shots have the same
                registers.
            strict_lengths: Whether to enforce that all register bitstrings have
                the same length.

        """
        shot_dct: dict[str, list[str]] = defaultdict(list)
        for shot in self.results:
            bitstrs = shot.to_register_bits()
            for reg, bitstr in bitstrs.items():
                if (
                    strict_lengths
                    and reg in shot_dct
                    and len(shot_dct[reg][0]) != len(bitstr)
                ):
                    msg = "All register bitstrings must have the same length."
                    raise ValueError(msg)
                shot_dct[reg].append(bitstr)
            if strict_names and bitstrs.keys() != shot_dct.keys():
                msg = "All shots must have the same registers."
                raise ValueError(msg)
        return dict(shot_dct)

    def to_pytket(self) -> BackendResult:
        """Convert results to a pytket BackendResult.

        Returns:
            BackendResult: A BackendResult object with the shots.

        Raises:
            ImportError: If pytket is not installed.
            ValueError: If a register's bitstrings have different lengths or not all
            registers are present in all shots.
        """
        try:
            from pytket._tket.unit_id import Bit
            from pytket.backends.backendresult import BackendResult
            from pytket.utils.outcomearray import OutcomeArray
        except ImportError as e:
            msg = "Pytket is an optional dependency, install with the `pytket` extra"
            raise ImportError(msg) from e
        reg_shots = self.register_bitstrings(strict_lengths=True, strict_names=True)
        reg_sizes: dict[str, int] = {
            reg: len(next(iter(reg_shots[reg]), "")) for reg in reg_shots
        }
        registers = list(reg_shots.keys())
        bits = [Bit(reg, i) for reg in registers for i in range(reg_sizes[reg])]

        int_shots = [
            int("".join(reg_shots[reg][i] for reg in registers), 2)
            for i in range(len(self.results))
        ]
        return BackendResult(
            shots=OutcomeArray.from_ints(int_shots, width=len(bits)), c_bits=bits
        )

    def _collated_shots_iter(self) -> Iterable[dict[str, list[DataValue]]]:
        for shot in self.results:
            yield shot.collate_tags()

    def collated_shots(self) -> list[dict[str, list[DataValue]]]:
        """For each shot generate a dictionary of tags to collated data."""
        return list(self._collated_shots_iter())

    def collated_counts(self) -> Counter[tuple[tuple[str, str], ...]]:
        """Calculate counts of bit strings for each tag by collating across shots using
        `QsysResult.tag_collated_shots`. Each `result` entry per shot is seen to be
        appending to the bitstring for that tag.

        If the result value is a list, it is flattened and appended to the bitstring.

        Example:
            >>> shots = [QsysShot([("a", 1), ("a", 0)]), QsysShot([("a", [0, 1])])]
            >>> res = QsysResult(shots)
            >>> res.collated_counts()
            Counter({(('a', '10'),): 1, (('a', '01'),): 1})

        Raises:
            ValueError: If any value is a float.
        """
        return Counter(
            tuple((tag, _flat_bitstring(data)) for tag, data in d.items())
            for d in self._collated_shots_iter()
        )


@deprecated("Use QsysResult instead.")
class HShots(QsysResult):
    """Deprecated alias for QsysResult."""


def _flat_bitstring(data: Iterable[DataValue]) -> str:
    return "".join(_cast_primitive_bit(prim) for prim in _flatten(data))


def _flatten(itr: Iterable[DataValue]) -> Iterable[DataPrimitive]:
    for i in itr:
        if isinstance(i, list):
            yield from _flatten(i)
        else:
            yield i



from hugr import Hugr, ops, tys
from hugr._serialization.serial_hugr import SerialHugr, serialization_version
from hugr.build.dfg import Dfg
from hugr.build.function import Module


def test_empty():
    h = SerialHugr(nodes=[], edges=[])
    assert h.model_dump() == {
        "version": serialization_version(),
        "nodes": [],
        "edges": [],
        "metadata": None,
        "encoder": None,
        "entrypoint": None,
    }


def test_children():
    mod = Module()
    mod.declare_function("foo", tys.PolyFuncType([], tys.FunctionType.empty()))

    h = mod.hugr
    assert len(h.children()) == 1

    h2 = Hugr.from_str(h.to_str())

    assert len(h2.children()) == 1


def test_entrypoint():
    dfg = Dfg(tys.Bool)
    noop = dfg.add_op(ops.Noop(tys.Bool), *dfg.inputs())
    dfg.set_outputs(noop)

    h = dfg.hugr
    assert len(h.children()) == 3
    assert h[noop].parent == dfg.to_node()

    func = h[dfg].parent
    assert h[func].op == ops.FuncDefn(
        f_name="main", inputs=[tys.Bool], _outputs=[tys.Bool]
    )
    assert h[func].parent == h.module_root

    # Do a roundtrip, and test all again
    h2 = Hugr.from_str(h.to_str())

    dfg = h2.entrypoint
    assert h2[dfg].op == ops.DFG(inputs=[tys.Bool], _outputs=[tys.Bool])
    assert len(h2.children()) == 3

    noop = h2.children(dfg)[2]
    assert h2[noop].parent == dfg.to_node()

    func = h2[dfg].parent
    assert h2[func].op == ops.FuncDefn(
        f_name="main", inputs=[tys.Bool], _outputs=[tys.Bool]
    )
    assert h2[func].parent == h2.module_root
    assert h2[h2.module_root].op == ops.Module()



from __future__ import annotations

import os
import pathlib
import subprocess
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import Self

from hugr import ext, tys
from hugr.envelope import EnvelopeConfig
from hugr.hugr import Hugr
from hugr.ops import AsExtOp, Command, DataflowOp, ExtOp, RegisteredOp
from hugr.package import Package
from hugr.std.float import FLOAT_T

if TYPE_CHECKING:
    from syrupy.assertion import SnapshotAssertion

    from hugr.ops import ComWire

QUANTUM_EXT = ext.Extension("pytest.quantum", ext.Version(0, 1, 0))
QUANTUM_EXT.add_op_def(
    ext.OpDef(
        name="H",
        description="Hadamard gate",
        signature=ext.OpDefSig(tys.FunctionType.endo([tys.Qubit])),
    )
)

QUANTUM_EXT.add_op_def(
    ext.OpDef(
        name="CX",
        description="CNOT gate",
        signature=ext.OpDefSig(tys.FunctionType.endo([tys.Qubit] * 2)),
    )
)


E = TypeVar("E", bound=Enum)


def _load_enum(enum_cls: type[E], custom: ExtOp) -> E | None:
    ext = custom._op_def._extension
    assert ext is not None
    if ext.name == QUANTUM_EXT.name and custom._op_def.name in enum_cls.__members__:
        return enum_cls(custom._op_def.name)
    return None


@dataclass(frozen=True)
class OneQbGate(AsExtOp):
    # Have to nest enum to avoid meta class conflict
    class _Enum(Enum):
        H = "H"

    _enum: _Enum

    def __call__(self, q: ComWire) -> Command:
        return DataflowOp.__call__(self, q)

    def op_def(self) -> ext.OpDef:
        return QUANTUM_EXT.operations[self._enum.value]

    @classmethod
    def from_ext(cls, custom: ExtOp) -> Self | None:
        return cls(e) if (e := _load_enum(cls._Enum, custom)) else None


H = OneQbGate(OneQbGate._Enum.H)


@dataclass(frozen=True)
class TwoQbGate(AsExtOp):
    class _Enum(Enum):
        CX = "CX"

    _enum: _Enum

    def op_def(self) -> ext.OpDef:
        return QUANTUM_EXT.operations[self._enum.value]

    @classmethod
    def from_ext(cls, custom: ExtOp) -> Self | None:
        return cls(e) if (e := _load_enum(cls._Enum, custom)) else None

    def __call__(self, q0: ComWire, q1: ComWire) -> Command:
        return DataflowOp.__call__(self, q0, q1)


CX = TwoQbGate(TwoQbGate._Enum.CX)


@QUANTUM_EXT.register_op(
    "Measure",
    signature=tys.FunctionType([tys.Qubit], [tys.Qubit, tys.Bool]),
)
@dataclass(frozen=True)
class MeasureDef(RegisteredOp):
    def __call__(self, q: ComWire) -> Command:
        return super().__call__(q)


Measure = MeasureDef()


@QUANTUM_EXT.register_op(
    "Rz",
    signature=tys.FunctionType([tys.Qubit, FLOAT_T], [tys.Qubit]),
)
@dataclass(frozen=True)
class RzDef(RegisteredOp):
    def __call__(self, q: ComWire, fl_wire: ComWire) -> Command:
        return super().__call__(q, fl_wire)


Rz = RzDef()


def _base_command() -> list[str]:
    workspace_dir = pathlib.Path(__file__).parent.parent.parent
    # use the HUGR_BIN environment variable if set, otherwise use the debug build
    bin_loc = os.environ.get("HUGR_BIN", str(workspace_dir / "target/debug/hugr"))
    return [bin_loc]


def mermaid(h: Hugr):
    """Render the Hugr as a mermaid diagram for debugging."""
    cmd = [*_base_command(), "mermaid", "-"]
    _run_hugr_cmd(h.to_str().encode(), cmd)


def validate(
    h: Hugr | Package,
    *,
    roundtrip: bool = True,
    snap: SnapshotAssertion | None = None,
):
    """Validate a HUGR or package.

    args:
        h: The HUGR or package to validate.
        roundtrip: Whether to roundtrip the HUGR through the CLI.
        snapshot: A hugr render snapshot. If not None, it will be compared against the
        rendered HUGR. Pass `--snapshot-update` to pytest to update the snapshot file.
    """
    # TODO: Use envelopes instead of legacy hugr-json
    cmd = [*_base_command(), "validate", "-"]

    serial = h.to_bytes(EnvelopeConfig.BINARY)
    _run_hugr_cmd(serial, cmd)

    if not roundtrip:
        return

    # Roundtrip checks
    if isinstance(h, Hugr):
        starting_json = h.to_str()
        h2 = Hugr.from_str(starting_json)
        roundtrip_json = h2.to_str()
        assert roundtrip_json == starting_json

        if snap is not None:
            dot = h.render_dot()
            assert snap == dot.source
            if os.environ.get("HUGR_RENDER_DOT"):
                dot.pipe("svg")
    else:
        # Package
        encoded = h.to_str(EnvelopeConfig.TEXT)
        loaded = Package.from_str(encoded)
        roundtrip_encoded = loaded.to_str(EnvelopeConfig.TEXT)
        assert encoded == roundtrip_encoded


def _run_hugr_cmd(serial: bytes, cmd: list[str]):
    """Run a HUGR command.

    The `serial` argument is the serialized HUGR to pass to the command via stdin.
    """
    try:
        subprocess.run(cmd, check=True, input=serial, capture_output=True)  # noqa: S603
    except subprocess.CalledProcessError as e:
        error = e.stderr.decode()
        raise RuntimeError(error) from e



"""HUGR model data structures."""

from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum
from typing import Protocol

import hugr._hugr as rust


class Term(Protocol):
    """A model term for static data such as types, constants and metadata."""

    def __str__(self) -> str:
        """Convert the term to its string representation."""
        return rust.term_to_string(self)

    @staticmethod
    def from_str(s: str) -> "Term":
        """Read the term from its string representation."""
        return rust.string_to_term(s)


@dataclass(frozen=True)
class Wildcard(Term):
    """Standin for any term."""


@dataclass(frozen=True)
class Var(Term):
    """Local variable, identified by its name."""

    name: str


@dataclass(frozen=True)
class Apply(Term):
    """Symbol application."""

    symbol: str
    args: Sequence[Term] = field(default_factory=list)


@dataclass(frozen=True)
class Splice:
    """A sequence spliced into the parent sequence."""

    seq: Term


SeqPart = Term | Splice


@dataclass(frozen=True)
class List(Term):
    """List of static data."""

    parts: Sequence[SeqPart] = field(default_factory=list)


@dataclass(frozen=True)
class Tuple(Term):
    """Tuple of static data."""

    parts: Sequence[SeqPart] = field(default_factory=list)


@dataclass(frozen=True)
class Literal(Term):
    """Static literal value."""

    value: str | float | int | bytes


@dataclass(frozen=True)
class Func(Term):
    """Function constant."""

    region: "Region"


@dataclass
class Param:
    """A parameter to a Symbol."""

    name: str
    type: Term

    def __str__(self):
        """Convert the parameter to its string representation."""
        return rust.param_to_string(self)

    @staticmethod
    def from_str(s: str) -> "Param":
        """Read the parameter from its string representation."""
        return rust.string_to_param(s)


@dataclass
class Symbol:
    """A named symbol."""

    name: str
    params: Sequence[Param] = field(default_factory=list)
    constraints: Sequence[Term] = field(default_factory=list)
    signature: Term = field(default_factory=Wildcard)

    def __str__(self):
        """Convert the symbol to its string representation."""
        return rust.symbol_to_string(self)

    @staticmethod
    def from_str(s: str) -> "Symbol":
        """Read the symbol from its string representation."""
        return rust.string_to_symbol(s)


class Op(Protocol):
    """The operation of a node."""


@dataclass(frozen=True)
class InvalidOp(Op):
    """Invalid operation intended to serve as a placeholder."""


@dataclass(frozen=True)
class Dfg(Op):
    """Dataflow graph."""


@dataclass(frozen=True)
class Cfg(Op):
    """Control flow graph."""


@dataclass(frozen=True)
class Block(Op):
    """Basic block in a control flow graph."""


@dataclass(frozen=True)
class DefineFunc(Op):
    """Function definiton."""

    symbol: Symbol


@dataclass(frozen=True)
class DeclareFunc(Op):
    """Function declaration."""

    symbol: Symbol


@dataclass(frozen=True)
class CustomOp(Op):
    """Custom operation."""

    operation: Term


@dataclass(frozen=True)
class DefineAlias(Op):
    """Alias definition."""

    symbol: Symbol
    value: Term


@dataclass(frozen=True)
class DeclareAlias(Op):
    """Alias declaration."""

    symbol: Symbol


@dataclass(frozen=True)
class TailLoop(Op):
    """Tail-controlled loop operation."""


@dataclass(frozen=True)
class Conditional(Op):
    """Conditional branch operation."""


@dataclass(frozen=True)
class DeclareConstructor(Op):
    """Constructor declaration."""

    symbol: Symbol


@dataclass(frozen=True)
class DeclareOperation(Op):
    """Operation declaration."""

    symbol: Symbol


@dataclass(frozen=True)
class Import(Op):
    """Import operation."""

    name: str


@dataclass
class Node:
    """A node in a hugr graph."""

    operation: Op = field(default_factory=lambda: InvalidOp())
    inputs: Sequence[str] = field(default_factory=list)
    outputs: Sequence[str] = field(default_factory=list)
    regions: Sequence["Region"] = field(default_factory=list)
    meta: Sequence[Term] = field(default_factory=list)
    signature: Term | None = None

    def __str__(self) -> str:
        return rust.node_to_string(self)


class RegionKind(Enum):
    """The kind of a hugr region."""

    DATA_FLOW = 0
    CONTROL_FLOW = 1
    MODULE = 2


@dataclass
class Region:
    """A hugr region containing an unordered collection of nodes."""

    kind: RegionKind = RegionKind.DATA_FLOW
    sources: Sequence[str] = field(default_factory=list)
    targets: Sequence[str] = field(default_factory=list)
    children: Sequence[Node] = field(default_factory=list)
    meta: Sequence[Term] = field(default_factory=list)
    signature: Term | None = None

    def __str__(self):
        """Convert the region to its string representation."""
        return rust.region_to_string(self)

    @staticmethod
    def from_str(s: str) -> "Region":
        """Read the region from its string representation."""
        return rust.string_to_region(s)


@dataclass
class Module:
    """A top level hugr module."""

    root: Region

    def __str__(self):
        """Convert the module to its string representation."""
        return rust.module_to_string(self)

    @staticmethod
    def from_str(s: str) -> "Module":
        """Read the module from its string representation."""
        return rust.string_to_module(s)


@dataclass
class Package:
    """A sequence of hugr modules."""

    modules: Sequence[Module]

    def __str__(self):
        """Convert the package into its string representation."""
        return rust.package_to_string(self)

    def __bytes__(self):
        """Convert the package into its binary representation."""
        return rust.package_to_bytes(self)

    @staticmethod
    def from_str(s: str) -> "Package":
        """Read a package from its string representation."""
        return rust.string_to_package(s)

    @staticmethod
    def from_bytes(b: bytes) -> "Package":
        """Read a package from its binary representation."""
        return rust.bytes_to_package(b)



import hugr.model

def term_to_string(term: hugr.model.Term) -> str: ...
def string_to_term(string: str) -> hugr.model.Term: ...
def node_to_string(node: hugr.model.Node) -> str: ...
def string_to_node(string: str) -> hugr.model.Node: ...
def region_to_string(region: hugr.model.Region) -> str: ...
def string_to_region(string: str) -> hugr.model.Region: ...
def param_to_string(region: hugr.model.Param) -> str: ...
def string_to_param(string: str) -> hugr.model.Param: ...
def symbol_to_string(region: hugr.model.Symbol) -> str: ...
def string_to_symbol(string: str) -> hugr.model.Symbol: ...
def module_to_string(module: hugr.model.Module) -> str: ...
def string_to_module(string: str) -> hugr.model.Module: ...
def module_to_bytes(module: hugr.model.Module) -> bytes: ...
def bytes_to_module(binary: bytes) -> hugr.model.Module: ...
def package_to_string(package: hugr.model.Package) -> str: ...
def string_to_package(string: str) -> hugr.model.Package: ...
def package_to_bytes(package: hugr.model.Package) -> bytes: ...
def bytes_to_package(binary: bytes) -> hugr.model.Package: ...



"""Value array types and operations."""

from __future__ import annotations

from dataclasses import dataclass

from hugr import tys, val
from hugr.std import _load_extension
from hugr.utils import comma_sep_str

EXTENSION = _load_extension("collections.value_array")


@dataclass(eq=False)
class ValueArray(tys.ExtType):
    """Fixed `size` array of `ty` elements that is copyable if its elements are."""

    def __init__(self, ty: tys.Type, size: int | tys.TypeArg) -> None:
        if isinstance(size, int):
            size = tys.BoundedNatArg(size)

        err_msg = (
            f"ValueArray size must be a bounded natural or a nat variable, not {size}"
        )
        match size:
            case tys.BoundedNatArg(_n):
                pass
            case tys.VariableArg(_idx, param):
                if not isinstance(param, tys.BoundedNatParam):
                    raise ValueError(err_msg)  # noqa: TRY004
            case _:
                raise ValueError(err_msg)

        ty_arg = tys.TypeTypeArg(ty)

        self.type_def = EXTENSION.types["value_array"]
        self.args = [size, ty_arg]

    @property
    def ty(self) -> tys.Type:
        assert isinstance(
            self.args[1], tys.TypeTypeArg
        ), "ValueArray elements must have a valid type"
        return self.args[1].ty

    @property
    def size(self) -> int | None:
        """If the value array has a concrete size, return it.

        Otherwise, return None.
        """
        if isinstance(self.args[0], tys.BoundedNatArg):
            return self.args[0].n
        return None

    def type_bound(self) -> tys.TypeBound:
        return self.ty.type_bound()


@dataclass
class ValueArrayVal(val.ExtensionValue):
    """Constant value for a statically sized value array of elements."""

    v: list[val.Value]
    ty: ValueArray

    def __init__(self, v: list[val.Value], elem_ty: tys.Type) -> None:
        self.v = v
        self.ty = ValueArray(elem_ty, len(v))

    def to_value(self) -> val.Extension:
        name = "VArrayValue"
        # The value list must be serialized at this point, otherwise the
        # `Extension` value would not be serializable.
        vs = [v._to_serial_root() for v in self.v]
        element_ty = self.ty.ty._to_serial_root()
        serial_val = {"values": vs, "typ": element_ty}
        return val.Extension(name, typ=self.ty, val=serial_val)

    def __str__(self) -> str:
        return f"value_array({comma_sep_str(self.v)})"



# Configuration file for the Sphinx documentation builder.  # noqa: INP001
# See https://www.sphinx-doc.org/en/master/usage/configuration.html
import hugr

project = "HUGR Python"
copyright = "2025, Quantinuum"
author = "Quantinuum"

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]

html_theme = "furo"


html_title = f"HUGR-py v{hugr.__version__} API documentation."

html_theme_options = {
    "sidebar_hide_name": False,
}

html_static_path = ["../_static"]

html_logo = "../_static/hugr_logo_no_bg.svg"

autosummary_generate = True

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "conftest.py"]


intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}

html_show_sourcelink = False



import pytest

from hugr import tys
from hugr.build.tracked_dfg import TrackedDfg
from hugr.package import Package
from hugr.std.float import FLOAT_T, FloatVal
from hugr.std.logic import Not

from .conftest import CX, QUANTUM_EXT, H, Measure, Rz, validate


def test_track_wire():
    dfg = TrackedDfg(tys.Bool, tys.Unit)
    inds = dfg.track_inputs()
    assert inds == [0, 1]
    assert dfg.tracked_wire(inds[0]) == dfg.inputs()[0]
    with pytest.raises(IndexError, match="Index 2 not a tracked wire."):
        dfg.tracked_wire(2)
    w1 = dfg.tracked_wire(inds[1])
    w1_removed = dfg.untrack_wire(inds[1])
    assert w1 == w1_removed
    with pytest.raises(IndexError, match="Index 1 not a tracked wire."):
        dfg.tracked_wire(inds[1])

    dfg.set_indexed_outputs(0)

    validate(dfg.hugr)


def simple_circuit(n_qb: int, float_in: int = 0) -> TrackedDfg:
    in_tys = [tys.Qubit] * n_qb + [FLOAT_T] * float_in
    return TrackedDfg(*in_tys, track_inputs=True)


def test_simple_circuit():
    circ = simple_circuit(2)
    circ.add(H(0))
    [_h, cx_n] = circ.extend(H(0), CX(0, 1))

    circ.set_tracked_outputs()

    assert len(circ.hugr) == 10

    # all nodes connected to output
    out_ins = {
        out.node
        for _, outs in circ.hugr.incoming_links(circ.output_node)
        for out in outs
    }
    assert out_ins == {cx_n}
    validate(Package([circ.hugr], [QUANTUM_EXT]))


def test_complex_circuit():
    circ = simple_circuit(2)
    fl = circ.load(FloatVal(0.5))

    circ.extend(H(0), Rz(0, fl))
    [_m0, m1] = circ.extend(*(Measure(i) for i in range(2)))

    m_idx = circ.track_wire(m1[1])  # track the bool out
    assert m_idx == 2
    circ.add(Not(m_idx))

    circ.set_tracked_outputs()

    assert len(circ.hugr) == 14

    assert circ._output_op().types == [tys.Qubit, tys.Qubit, tys.Bool]

    validate(Package([circ.hugr], [QUANTUM_EXT]))
